import sys
from scipy import special, stats
from numpy import array as A

def compoundPartitioning(agents):
    """Compute and return sections with compound criteria

    agents is a dict with keys "d", "id", "od", "s", "is", "os"
    with sectorialized_agents__ with each of these criteria
    """
    exc_h=set( agents["d"][-1][2]) & \
          set(agents["id"][-1][2]) & \
          set(agents["od"][-1][2]) & \
          set( agents["s"][-1][2]) &  \
          set(agents["is"][-1][2]) & \
          set(agents["os"][-1][2])
    exc_i=set( agents["d"][-1][1]) & \
          set(agents["id"][-1][1]) & \
          set(agents["od"][-1][1]) & \
          set( agents["s"][-1][1]) & \
          set(agents["is"][-1][1]) & \
          set(agents["os"][-1][1])
    exc_p=set( agents["d"][-1][0]) & \
          set(agents["id"][-1][0]) & \
          set(agents["od"][-1][0]) & \
          set( agents["s"][-1][0]) & \
          set(agents["is"][-1][0]) & \
          set(agents["os"][-1][0])
    exc=exc_p,exc_i,exc_h

    inc_h=set( agents["d"][-1][2])  | \
          set(agents["id"][-1][2]) | \
          set(agents["od"][-1][2]) | \
          set( agents["s"][-1][2])  | \
          set(agents["is"][-1][2]) | \
          set(agents["os"][-1][2])
    inc_i=set( agents["d"][-1][1]) | \
          set(agents["id"][-1][1]) | \
          set(agents["od"][-1][1]) | \
          set( agents["s"][-1][1]) | \
          set(agents["is"][-1][1]) | \
          set(agents["os"][-1][1])
    inc_p=set( agents["d"][-1][0]) | \
          set(agents["id"][-1][0]) | \
          set(agents["od"][-1][0]) | \
          set( agents["s"][-1][0]) | \
          set(agents["is"][-1][0]) | \
          set(agents["os"][-1][0])
    inc=inc_p, inc_i, inc_h

    total=set(agents["d"][-1][0]+agents["d"][-1][1]+agents["d"][-1][2])
    excc_h=exc[2]
    excc_p=inc[0]
    #excc_i=total - (exc[2] & inc[0])
    excc_i=total - (exc[2] | inc[0])
    excc=excc_p,excc_i,excc_h

    incc_h=inc[2]
    incc_p=excc[0]
    incc_i=total-(incc_h | incc_p)
    incc=incc_p,incc_i,incc_h

    exce_h=exc[2]
    exce_i=inc[1]
    exce_p=total-(exce_h | exce_i)
    exce=exce_p,exce_i,exce_h

    ince_h=inc[2]
    ince_i=exc[1]
    ince_p=total-(ince_h | ince_i)
    ince=ince_p,ince_i,ince_h

    return dict(total=total, exc=exc, inc=inc, excc=excc, incc=incc, exce=exce, ince=ince)



class NetworkPartitioning:
    network_count=0
    def __init__(self,networkMeasures=None, minimum_incidence=1,metric="strength"):
        if not networkMeasures:
            networkMeasures=g.NetworkMeasures()
        self.metric=metric
        metric_=self.standardizeName(metric)

        prob, max_degree_empirical, max_degree_possible = \
                self.basicMeasures( networkMeasures , metric_)

        incident_degrees, incident_degrees_, agent_degrees = \
                  self.makeDegreeLists( networkMeasures, metric_)

        empirical_distribution = self.makeEmpiricalDistribution(
            incident_degrees, incident_degrees_, networkMeasures.N )

        binomial_distribution=self.makeBinomialDistribution(
                   prob, max_degree_possible, incident_degrees_)

        binomial=stats.binom(max_degree_possible,prob)

        #sectorialized_degrees= self.sectorializeDegrees(
        # empirical_distribution, binomial_distribution, incident_degrees_)

        #sectorialized_degrees_= self.newSectorializeDegrees(
        # empirical_distribution, binomial_distribution, incident_degrees_)

        sectorialized_degrees__= self.newerSectorializeDegrees(
              empirical_distribution, binomial, incident_degrees_,
              max_degree_empirical,minimum_incidence,networkMeasures.N )

        #sectorialized_agents= self.sectorializeAgents(
        #     sectorialized_degrees, networkMeasures.degrees)

        #sectorialized_agents_= self.sectorializeAgents(
        #     sectorialized_degrees_, networkMeasures.degrees)

        sectorialized_agents__= self.sectorializeAgents(
             sectorialized_degrees__, agent_degrees)

        NetworkPartitioning.network_count+=1 # to keep track of how may partitions have been done

        self.makeSelf("incident_degrees_     ",incident_degrees_     ,
                      "incident_degrees     ",incident_degrees     ,
                      #"sectorialized_agents  ",sectorialized_agents  ,
                      #"sectorialized_agents_  ",sectorialized_agents_  ,
                      "sectorialized_agents__  ",sectorialized_agents__  ,
                      #"sectorialized_degrees ",sectorialized_degrees ,
                      #"sectorialized_degrees_ ",sectorialized_degrees_ ,
                      "sectorialized_degrees__ ",sectorialized_degrees__ ,
                      "binomial_distribution ",binomial_distribution ,
                      "prob"                  ,prob,
                      "max"                   ,(max_degree_possible, max_degree_empirical),
                      "empirical_distribution",empirical_distribution,
                      "binomial",binomial,
                      "metric_",metric_,
                      "minimum_incidence",minimum_incidence,
                      "binomial_distribution" ,binomial_distribution)

    def makeSelf(self, *args):
        for signifier, signified  in zip(args[::2], args[1::2]):
            #try:
                exec("self.{} = signified".format(signifier))
                #thing=signified
                #exec("self.{} = thing".format(signifier))
                #exec("self.{} = {}".format(signifier, signified))
                #exec("self.{} = ".format(signifier), signified)
            #except:
            #    self.binomial=signified
    def standardizeName(self,name):
        if name in (["s","strength","st"]+["f","força","forca","fo"]):
            name_="s"
        elif name in (["is","in_strength","ist"]+["fe","força_e","forca_e","fe"]):
            name_="is"
        elif name in (["os","out_strength","ost"]+["fs","força_s","forca_s","fs"]):
            name_="os"
        elif name in (["d","degree","dg"]+["g","grau","gr"]):
            name_="d"
        elif name in (["id","in_degree","idg"]+["ge","grau_e","gre"]):
            name_="id"
        elif name in (["od","out_degree","odg"]+["gs","grau_s","grs"]):
            name_="od"
        return name_

    def basicMeasures(self,networkMeasures,metric_):
        nm=networkMeasures
        if metric_ in ("s","is","os"):
            edge_weights=[i[2]["weight"] for i in nm.edges]
            average_edge_weight=sum(edge_weights)/nm.E
            self.average_edge_weight=average_edge_weight
        if metric_=="s":
            max_degree_empirical=round(max(nm.strengths_) / average_edge_weight)
        elif metric_=="is":
            max_degree_empirical=round(2*max(nm.in_strengths_) / average_edge_weight)
        elif metric_=="os":
            max_degree_empirical=round(2*max(nm.out_strengths_) / average_edge_weight)
        elif metric_=="d":
            max_degree_empirical=max(nm.degrees_)
        elif metric_=="id":
            max_degree_empirical=2*max(nm.in_degrees_)
        elif metric_=="od":
            max_degree_empirical=2*max(nm.out_degrees_)
        prob=nm.E/(nm.N*(nm.N-1)) # edge probability
        max_degree_possible=2*(nm.N-1) # max d given N
        return prob, max_degree_empirical, max_degree_possible
    def makeDegreeLists(self, networkMeasures,metric_):
        if metric_=="s":
            agent_degrees={i:round(j/self.average_edge_weight) for i,j in networkMeasures.strengths.items()}
            incident_degrees=list(agent_degrees.values())
        elif metric_=="is":
            agent_degrees={i:round((2*j)/self.average_edge_weight) for i,j in networkMeasures.in_strengths.items()}
            incident_degrees=list(agent_degrees.values())
        elif metric_=="os":
            agent_degrees={i:round((2*j)/self.average_edge_weight) for i,j in networkMeasures.out_strengths.items()}
            incident_degrees=list(agent_degrees.values())
        elif metric_=="d":
            agent_degrees=networkMeasures.degrees
            incident_degrees=networkMeasures.degrees_
        elif metric_=="id":
            agent_degrees={i:(2*j) for i,j in networkMeasures.in_degrees.items()}
            incident_degrees=list(agent_degrees.values())
        elif metric_=="od":
            agent_degrees={i:(2*j) for i,j in networkMeasures.out_degrees.items()}
            incident_degrees=list(agent_degrees.values())
        incident_degrees_=list(set(incident_degrees))
        incident_degrees_.sort()
        return incident_degrees, incident_degrees_, agent_degrees
    def makeEmpiricalDistribution(self, incident_degrees, incident_degrees_, N):
        empirical_distribution=[]
        for degree in incident_degrees_:
            empirical_distribution.append(incident_degrees.count(degree)/N)
        return empirical_distribution
    def makeBinomialDistribution(self,prob,max_degree_possible,incident_degrees_):
        """If max_degree_possible == max_degree_empirical, makeBinomial ==1"""
        binomial_distribution=[] # occurance probability of degrees 
        for degree in incident_degrees_:
            if len(binomial_distribution) and binomial_distribution[-1]==0.0:
                binomial_distribution.append(0.0)
            else:
                n_occurrences=special.binom(max_degree_possible,degree)
                prob_degree=n_occurrences *  (prob**degree)*((1-prob)**(max_degree_possible-degree))
                binomial_distribution.append(prob_degree)
        return binomial_distribution

    def sectorializeAgents(self,sectorialized_degrees,agent_degrees):
        periphery=[x for x in agent_degrees
                     if agent_degrees[x] in sectorialized_degrees[0]]
        intermediary=[x for x in agent_degrees
                     if agent_degrees[x] in sectorialized_degrees[1]]
        hubs=[x for x in agent_degrees
                     if agent_degrees[x] in sectorialized_degrees[2]]
        return periphery, intermediary, hubs

    def newerSectorializeDegrees(self,empirical_distribution,binomial,incident_degrees_,max_degree_empirical,minimum_count,num_agents):
        # compute bins [start, end]
        prob_min=minimum_count/num_agents
        llimit=0
        rlimit=0
        self.bins=bins=[]
        self.empirical_probs=empirical_probs=[]
        while (rlimit < len(incident_degrees_)):
            if (sum(empirical_distribution[llimit:])>prob_min):
                prob_empirical=0
                while True:
                    prob_empirical=sum(
                         empirical_distribution[llimit:rlimit+1] )
                    if prob_empirical >= prob_min:
                        break
                    else:
                        rlimit+=1
                bins.append((llimit,rlimit))
                empirical_probs.append(prob_empirical)
                rlimit+=1
                llimit=rlimit
            else: # last bin
                print("last bin less probable than prob_min")
                rlimit=len(incident_degrees_)-1
                bins.append((llimit,rlimit))
                prob_empirical=sum(
                     empirical_distribution[llimit:rlimit+1] )
                empirical_probs.append(prob_empirical)
                rlimit+=1

        binomial_probs=[]
        for i, bin_ in enumerate(bins):
            llimit=bin_[0]
            rlimit=bin_[1]
            ldegree=incident_degrees_[llimit]-1
            rdegree=incident_degrees_[rlimit]
            binomial_prob=binomial.cdf(rdegree)-binomial.cdf(ldegree)
            binomial_probs.append(binomial_prob)

        # calcula probabilidades em cada bin
        # compara as probabilidades
        distribution_compare = list(A(empirical_probs) < A(binomial_probs))
        self.binomial_probs=binomial_probs
        self.distribution_compare0=distribution_compare
        if sum(distribution_compare):
            tindex= distribution_compare.index(True)
            tindex2=distribution_compare[::-1].index(True)
            periphery_degrees=incident_degrees_[:tindex]
            intermediary_degrees=incident_degrees_[tindex:-tindex2]
            hub_degrees=         incident_degrees_[-tindex2:]
        else:
            periphery_degrees=incident_degrees_[:]
            intermediary_degrees=[]
            hub_degrees=[]

        return periphery_degrees, intermediary_degrees, hub_degrees

    def newSectorializeDegrees(self,empirical_distribution,binomial_distribution,incident_degrees_):
        distribution_compare = A(empirical_distribution) < A(binomial_distribution)
        self.distribution_compare=distribution_compare
        tindex= list(distribution_compare      ).index(True)
        tindex2=list(distribution_compare[::-1]).index(True)
        periphery_degrees=incident_degrees_[:tindex]
        intermediary_degrees=incident_degrees_[tindex:-tindex2]
        hub_degrees=         incident_degrees_[-tindex2:]
        return periphery_degrees, intermediary_degrees, hub_degrees
    def sectorializeDegrees(self,empirical_distribution,binomial_distribution,incident_degrees_):
        periphery_degrees=[]
        intermediary_degrees=[]
        hub_degrees=[]
        lock=0
        lock2=0
        for incident_prob, binomial_prob, degree in zip(
  empirical_distribution, binomial_distribution, incident_degrees_):
            if incident_prob < binomial_prob:
                intermediary_degrees.append(degree)
                lock=1
            elif (incident_prob > binomial_prob) and lock:
                hub_degrees.append(degree)

            else:
                periphery_degrees.append(degree)
        return periphery_degrees, intermediary_degrees, hub_degrees



from .downloadMessages import DownloadGmaneData
import mailbox, os, percolation as P
c=P.check

class LoadMessages:
    """Class that loads Gmane messages saved locally
    
    Usage
    =====
    After downloading messages from Gmane lists with the 
    DownloadGmaneData class, see chosen basedir,
    or DownloadGmaneData.downloadedLists after
    DownloadGmaneData.getDownloadedLists()
    or basedir/stats.txt after DownloadGmaneData.downloadedStats()
    to select the list_id
    
    Then go
    LoadMessages(list_id,100)
    to load first 100 messages or see the initialization of such class
    if n_messages=-1, class loads all messages (default)

    OBS: any message M on self.messagesB have both M[0] and M[1]
    (and might have more items).
    Only M[0] is on self.messages, which is used for analysis.
    All M[1] were found empty to date, but it is wise to check.
    """
    def __init__(self,list_id=None,n_messages=None,offset=0,basedir="~/.gmane/",loggin_file="load.log"):
        self._BASE_DIR=basedir.replace("~",os.path.expanduser("~"))
        if not os.path.isdir(self._BASE_DIR):
            os.mkdir(self._BASE_DIR)
        if not list_id:
            dl=DownloadGmaneData() # saves into ~/.gmane/
            dl.getDownloadedLists()
            list_id=dl.downloaded_lists[0]
        self.list_id=list_id
        self.n_messages=n_messages
        self.offset=offset
        self.n_empty=0
        self.loadMessages()
    def loadMessages(self):
        mfiles=os.listdir(self._BASE_DIR+self.list_id)
        mfiles.sort()
        messages=[]
        messagesB=[]
        if self.n_messages:
            tend=self.offset+self.n_messages
        else:
            tend=None
        mcount=0
        for mfile in mfiles[self.offset:tend]:
            mcount+=1
            if mcount%300==0: c("load +300: {}".format(mcount))
            mbox = mailbox.mbox(self._BASE_DIR+self.list_id+"/"+mfile)
            if mbox.keys():
                messages.append(mbox[0])
                if len(mbox.keys())>1:
                    messagesB.append(mbox)
            else:
                self.n_empty+=1
        self.messages=messages
        self.messagesB=messagesB
        if not self.n_messages:
            self.n_messages=len(messages)



from .downloadMessages import *
__all__=["DownloadGmaneData",
"LoadMessages",
"ListDataStructures",
"InteractionNetwork",
"NetworkMeasures",
"compoundPartitioning", "NetworkPartitioning",
"NetworkDrawer",
"NetworkEvolution",
"NetworkPCA",
"circularStatistics", "TimeStatistics",
"AgentStatistics",
"EvolutionTimelines",
"makeTables", "partialSums", "pcaTable", "writeTex",
"EvolutionMusic",]
#"utils",
#"EmailStructures"]



from .downloadMessages import DownloadGmaneData
from .loadMessages import LoadMessages
from .listDataStructures import ListDataStructures
from .interactionNetwork import InteractionNetwork
from .networkMeasures import NetworkMeasures
from .networkPartitioning import compoundPartitioning, NetworkPartitioning
from .networkDrawer import NetworkDrawer
from .networkEvolution import NetworkEvolution
from .pca import NetworkPCA
from .timeStatistics import circularStatistics, TimeStatistics
from .agentStatistics import AgentStatistics
from .evolutionTimelines import EvolutionTimelines
from . import tableHelpers
from .tableHelpers import *
#from .tableHelpers import makeTables, partialSums, pcaTable, writeTex, markEntries, lTable, encapsulateTable, markEntries_, doubleLines, fSize
from .evolutionMusic import EvolutionMusic
from .textUtils import EmailStructures, generalMeasures, makeText, medidasLetras, medidasTokens, medidasTamanhosTokens, medidasTamanhosSentencas, medidasTamanhosMensagens, medidasPOS, medidasWordnet, medidasWordnet2, tfIdf, WL_,medidasTokensQ,medidasTokensQ_, perc, digRoot, makeGeneralTable
from .ksStatistics import KSReferences, kolmogorovSmirnovDistance, kolmogorovSmirnovDistance_

from . import ksStatistics as KS
from . import textUtils
from . import utils
from . import triplifyList




import pickle, os
from .listDataStructures import *
from .interactionNetwork import *
from .networkMeasures import *
from .networkPartitioning import *
from .networkDrawer import *
from .agentStatistics import *
from .timeStatistics import *
from .pca import *
import time as TT
class NetworkEvolution:
    def __init__(self, window_size=200, step_size=3, make_analysis=True, write_analysis=True, make_pca=True, write_pca=True,tdir="evolution"):
        self.tdir=tdir
        self.window_size=window_size
        self.step_size=step_size
        self.make_analysis=make_analysis
        self.write_analysis=write_analysis
        self.make_pca      =make_pca
        self.write_pca     =write_pca
    def evolveDataStructuresTimeline(self,list_structures):
        """Use snapshots to make network timelines"""
        pointer=0
        class list_structures_:
            messages=list_structures.messages
        self.interaction_networks=[]
        self.networks_measures=[]
        self.networks_partitionings=[]
        while pointer+self.window_size<list_structures.n_messages:
            T=TT.time()
            list_structures_.message_ids=list_structures.message_ids[pointer:pointer+self.window_size]
            iN=InteractionNetwork(list_structures_)
            nm=NetworkMeasures(iN,exclude=[
#                "weighted_directed_betweenness",
                "unweighted_directed_betweenness",
                "weighted_undirected_betweenness",
                "unweighted_undirected_betweenness",
                "weiner",
                "closeness",
                "transitivity",
                "rich_club",
#                "weighted_clustering",
                "triangles",
                "n_weakly_connected_components",
                "n_strongly_connected_components", 
                ])
            np=NetworkPartitioning(nm,minimum_incidence)
            del np.binomial
            self.interaction_networks.append(iN)
            self.networks_measures.append(nm)
            self.networks_partitionings.append(np)
            with open("{}/im{:09}.pickle".format(tdir,counter),"wb") as f:
                tall=dict(ds=ds,ts=ts,ps=ps,iN=iN,nm=nm,
                        agents=np.sectorialized_agents__,
                        minimum_incidence=np.minimum_incidence,
                        np=np)
                pickle.dump(tall,f)
            print("ws: {}, pointer: {}, ss: {}, took: {:.2f}".format(
                self.window_size, pointer,self.step_size,TT.time()-T))
            pointer+=self.step_size

    def evolveDataStructures(self,list_structures,minimum_incidence=2):
        """Use snapshots to make networks and PCA of each"""

        # make network from data_structures messages
        counter=0
        pointer=0
        class list_structures_:
            messages=list_structures.messages
        self.interaction_networks=[]
        self.networks_measures=[]
        self.networks_pcas=[]
        while pointer+self.window_size<list_structures.n_messages:
            T=TT.time()
            list_structures_.message_ids=list_structures.message_ids[pointer:pointer+self.window_size]
            iN=InteractionNetwork(list_structures_)
            nm=NetworkMeasures(iN,exclude=[
#                "weighted_directed_betweenness",
                "unweighted_directed_betweenness",
                "weighted_undirected_betweenness",
                "unweighted_undirected_betweenness",
                "weiner",
                "closeness",
                "transitivity",
                "rich_club",
#                "weighted_clustering",
                "triangles",
                "n_weakly_connected_components",
                "n_strongly_connected_components", 
                ])
            npca=NetworkPCA(nm,None)
            self.interaction_networks.append(iN)
            self.networks_measures.append(nm)
            self.networks_pcas.append(npca)
            np=NetworkPartitioning(nm,minimum_incidence)
            del np.binomial
            if self.tdir: os.system("mkdir {}".format(self.tdir))
            with open("{}/im{:09}.pickle".format(self.tdir,counter),"wb") as f:
                tall=dict(nm=nm,
                        agents=np.sectorialized_agents__,
                        minimum_incidence=np.minimum_incidence,
                        np=np,
                        npca=npca)
                pickle.dump(tall,f)

            print("ws: {}, pointer: {}, ss: {}, took: {:.2f}".format(
                self.window_size, pointer,self.step_size,TT.time()-T))
            pointer+=self.step_size
            counter+=1
        # of chunks of size window_size
        self.offset=0
        self.minimum_incidence=minimum_incidence
        ds=list_structures
        #ts=TimeStatistics(ds)
        #ps=AgentStatistics(ds)
        iN=InteractionNetwork(ds)
        nm=NetworkMeasures(iN,exclude=[
#                "weighted_directed_betweenness",
                "unweighted_directed_betweenness",
                "weighted_undirected_betweenness",
                "unweighted_undirected_betweenness",
                "weiner",
                "closeness",
                "transitivity",
                "rich_club",
#                "weighted_clustering",
                "triangles",
                "n_weakly_connected_components",
                "n_strongly_connected_components", 
                ])
        np=NetworkPartitioning(nm,3)

        #self.overall=(ds,ps,iN,nm,np)
        self.overall=(ds,iN,nm,np)
        self.saveOverallInfo()
        # and separation step_size
        # make pca of each
        # keep both PCAs and networks in class variables and pickle files

    def evolveRaw(self, loaded_messages,offset=0,tdir="evolution",clean_dir=True,print_status=True,imagerate=18,erdos_sectors=True,minimum_incidence=2):
        if len(loaded_messages)<self.window_size:
            raise ValueError("incoming messages smaller than window_size [({}-{})/{}]".format(
                len(loaded_messages), offset, self.window_size))
        pointer=offset
        self.offset=offset
        self.minimum_incidence=minimum_incidence
        self.tdir=tdir
        counter=0

        if tdir: os.system("mkdir {}".format(tdir))
        if clean_dir: os.system("rm {}/*".format(tdir))
        while pointer+self.window_size<len(loaded_messages):
            messages=loaded_messages[pointer:pointer+self.window_size]

            ds=ListDataStructures(messages)
            ts=TimeStatistics(ds)
            ps=AgentStatistics(ds)
            iN=InteractionNetwork(ds)
            nm=NetworkMeasures(iN)
            np=NetworkPartitioning(nm,minimum_incidence)
            npca=NetworkPCA(nm,np,tdir=tdir,tname="pca{:09}".format(counter))
            del np.binomial
            with open("{}/im{:09}.pickle".format(tdir,counter),"wb") as f:
                tall=dict(ds=ds,ts=ts,ps=ps,iN=iN,nm=nm,
                        agents=np.sectorialized_agents__,
                        minimum_incidence=np.minimum_incidence,
                        np=np,
                        npca=npca)
                pickle.dump(tall,f)
            if "drawer" not in dir(self):
                print("runing self.setDrawer to enable images and movie")
                self.setDrawer(loaded_messages,tdir=tdir)
            label="T{}W{}S{}R{}M{}N{}E{}".format(
                len(loaded_messages), self.window_size, self.step_size, 
                imagerate, pointer,nm.N,nm.E)
            self.drawer.drawNetwork( iN,nm ,"{}/im{:09}.png".format(tdir,counter), label,np)
            if print_status:
                print("analysed {}-{} / {}".format(
                    pointer,pointer+self.window_size,len(loaded_messages)))
            pointer+=self.step_size
            counter+=1
            # make pca 
            # write pca with cpickle
        self.saveOverallInfo()
        videoname="evoT{}W{}O{}S{}R{}".format(
                len(loaded_messages), self.window_size, offset, self.step_size, imagerate)
        self.makeVideo(tdir,videoname,imagerate)
    def saveOverallInfo(self):
        with open("{}/overallInfo.pickle".format(self.tdir),"wb") as f:
            tdict=dict( window_size=self.window_size,
                         step_size =self.step_size,
                            offset =self.offset,
                  minimum_incidece =self.minimum_incidence )
            del self.overall[-1].binomial
            pickle.dump((tdict,self.overall),f)

    def makeVideo(self,tdir="evolution",videoname="fooname",imagerate=18):
        command="avconv -f image2 -framerate {} -i {}/im%09d.png -qscale 1 -b 65536k -y {}/{}.avi".format(
                imagerate, tdir, tdir, videoname)
        os.system(command)
    def setDrawer(self,messages=None,network_measures=None, network_partitioning=None,tdir="."):
        if messages:
            ds=ListDataStructures(messages)
            ts=TimeStatistics(ds)
            ps=AgentStatistics(ds)
            iN=InteractionNetwork(ds)
            nm=NetworkMeasures(iN)
            np=NetworkPartitioning(nm,3)
        else:
            print("under construction")
        drawer=NetworkDrawer()
        drawer.makeLayout(nm,np)
        drawer.drawNetwork(iN,nm,"{}/overall.png".format(tdir))
        drawer.step_size=self.step_size
        drawer.offset=self.offset
        drawer.window_size=self.window_size
        self.drawer=drawer
        self.drawer=drawer
        self.overall=(ds,ts,ps,iN,nm,np)




import numpy as n, string, re, builtins as B, langid
def doubleColumn(tablefname):
    with open(tablefname,"r") as f:
        lines=f.read()
    l=lines.replace("\\begin{table}","\\begin{table*}").replace("\\end{table}","\\end{table*}")
    writeTex(l,tablefname)

def fSize(tablefname,ftag="scriptsize",write=False):
    """Change size of table font"""
    with open(tablefname,"r") as f:
        lines=f.read()
    l=lines.split("\n")
    l.insert(1,ftag)
    l="\n".join(l)
    if not write:
        return l
    else:
        writeTex(l,tablefname)


def doubleLines(tablefname,hlines=["i1","i2"],vlines=["j1","j2"],hlines_=[]):
    """make some double lines in a latex table"""
    with open(tablefname,"r") as f:
        lines=f.read()
    # colocando barras nas linhas verticais
    header=re.findall(r"\\begin{tabular}{(.*)}\\hline\n",lines)[0]
    indexes=[i.start() for i in re.finditer("\|",header)]
    js=vlines
    header_=header[:]
    foo=0
    for j in js:
        j_=indexes[j]+foo
        header_=header_[:j_]+"||"+header_[j_+1:]
        foo+=1
    lines__=lines.replace(header,header_)
    # colocando barras nas linhas horizontais
    linhas=lines__.split("\\hline")
    ii=hlines
    linesF=lines__[:]
    for i in ii:
        linha=linhas[i]
        linha_=linha+"\\hline"
        if lines__.count(linha)==1:
            linesF=linesF.replace(linha,linha_)
        elif lines__.count(linha)>1:
            print("mais de uma linha igual ERRO!!!",tablefname)
        else:
            print("linha não existe!!! ERRO!!",tablefname)
    ii=hlines_
    for i in ii:
        linha=linhas[i]
        if lines__.count(linha)==1:
            linesF=linesF.replace(linha+"\\hline",linha)
        elif lines__.count(linha)>1:
            print("mais de uma linha igual ERRO!!!",tablefname)
        else:
            print("linha não existe!!! ERRO!!",tablefname)
    #linesF=lines__[:]
    return linesF
def markEntries_(tablefname,marker,locs=[("i","j")]):
    with open(tablefname,"r") as f:
        lines=f.read()
    for loc in locs:
        linha=re.findall(r"\\hline\n.*"*loc[0]+r"\\hline\n(.*)\\hline\n",lines)[0]

        elementos=[i.strip() for i in linha.split("&")]
        elementos_=elementos[:]
        elementos_[-1],resto=elementos[-1].split(" ")
        elemento=elementos_[loc[1]]
        elemento_="{{{} {}}}".format(marker,elemento)
        elementos_[loc[1]]=elemento_
        linha_=" & ".join(elementos_)+" "+resto
        if lines.count(linha)==1:
            #print(linha,linha_)
            lines=lines.replace(linha,linha_)
        elif lines.count(linha)>1:
            print("mais de uma linha igual ERRO!!!",tablefname)
        else:
            print("linha não existe!!! ERRO!!",tablefname)
    #print("\n\n\nLINES",lines)
    return lines
def markEntries(table,marker):
    """Make entries in a table boldface or use other markings"""
    # open rendered table as text
    #print("YEAH")
    with open(table,"r") as f:
        lines=f.read()
    lines=lines.split("\n")
    lines[3:-3]=[i.split("&") for i in lines[3:-3]]
    for i,line in enumerate(lines[3:-3]):
        i1,i2=line[-1].split("\\\\\\")
        i2="\\\\\\"+i2
        lines[i+3]=line[:-1]+[i1,i2]
    #print(lines)
    #print(len(lines[0]))
    for column in range(1,len(lines[3])-1):
        #print("column")
        values=[]
        for linei in range(3,len(lines)-3):
            if lines[linei][column].strip(): # can be empty
                value=float(lines[linei][column].split("{")[-1].split("}")[0])
                values.append(value)
        #print(values)
        mav=max(values)
        miv=min(values)
        #print(mav,miv)
        for linei in range(3,len(lines)-3):
            if lines[linei][column].strip(): # can be empty
                orig="{:.2f}".format(mav)
                tnew="\\{}{{ {} }}".format(marker,mav)
                #print(orig,tnew)
                lines[linei][column]=lines[linei][column].replace(orig,tnew)
    lines=lines[:3]+[" & ".join(line[:-1])+line[-1] for line in lines[3:-3]]+lines[-3:]
    lines=" \n ".join(lines)
    writeTex(lines,table.replace(".tex","_.tex"))
    # find maximum and minimum in each row
    # boldface them all
    return lines
def encapsulateTable(string_table,column_labels, caption,ttype=None):
    """Uses the output of makeTables to render a complete latex table"""
    if ttype in ("kolmDiff3","kolmDiff3_","kolmSamp_","kolmSamp","textCorr","allFloat"):
        header="\\begin{table*}[h!]\n\\begin{center}\n\\begin{tabular}{| l |"+" c |"*(string_table.split("hline")[0].count("&")) +"}\\hline\n"
    elif ttype=="audioDistances":
        header="\\begin{table*}[h!]\n\\begin{center}\n\\begin{tabular}{| l |"+" c |"*(string_table.split("hline")[0].count("&")) +"}\\hline\n"
    elif ttype=="textsDistances":
        header="\\begin{table*}[h!]\n\\begin{center}\n\\begin{tabular}{| l |"+" c |"*(string_table.split("hline")[0].count("&")) +"}\\hline\n"
    elif ttype=="textsGeneral":
        header="\\begin{table*}[h!]\n\\begin{center}\n\\begin{tabular}{| l |"+" c |"*(string_table.split("hline")[0].count("&")) +"}\\hline\n"
    else:
        header="\\begin{table}[h!]\n\\begin{center}\n\\begin{tabular}{| l |"+" c |"*(string_table.split("hline")[0].count("&")) +"}\\hline\n"
#    header+="\\footnotesize"
    if column_labels:
        header+=("& {} "*len(column_labels)+"\\\\\\hline\n").format(*column_labels)[2:]
    if "tag" in dir(B):
        caption+=" TAG: {}".format(B.tag)
    caption_="\\caption{{{}}}\n".format(caption)
    if ttype in ("kolmDiff3","kolmDiff3_","kolmSamp_","kolmSamp","textCorr","allFloat"):
        footer="\\end{{tabular}}\n{}\\end{{center}}\n\\end{{table*}}".format(caption_)
    elif ttype=="audioDistances":
        footer="\\end{{tabular}}\n{}\\end{{center}}\n\\end{{table*}}".format(caption_)
    elif ttype=="textsGeneral":
        footer="\\end{{tabular}}\n{}\\end{{center}}\n\\end{{table*}}".format(caption_)
    elif ttype=="textsDistances":
        footer="\\end{{tabular}}\n{}\\end{{center}}\n\\end{{table*}}".format(caption_)
    else:
        footer="\\end{{tabular}}\n{}\\end{{center}}\n\\end{{table}}".format(caption_)
    table=header+string_table+footer
    return table
def lTable(labels,labelsh,data,caption,filename,ttype="kolmNull",two_decimal=True):
    t1=makeTables(labels,data,two_decimal,ttype)
    t1_=t1.split("\\hline")[:-1]
    #tvals=[int(float(tt.split("&")[-1].split(" \\\\")[0])) for tt in t1_]
    #print("AQUIIII!!!B",tvals)
    #print(t1)
    #for tval in tvals:
    #    print(str(tval)+".00",str(tval))
    #    t1=t1.replace(" {}.00 ".format(tval)," {} ".format(tval))
    t2=encapsulateTable(t1,labelsh,caption,ttype)
    B.b=(t1,t1_,t2)
    writeTex(t2,filename)
#def makeTables2(data,two_decimal=False):
#    """Variation of makeTables for tables withour row labels"""
def makeTables(labels,data,two_decimal=False,ttype=None):
    """Returns a latex table of data with Label in first column.
    
    Returns latex printable or writable to files to
    be imported by latex files.
    ttype gives a known table type:
    kolmNull for the Kolmogorov-Smirnov Null hypothesis checking.
    """
    if len(labels)!=len(data):
        print("input one label per data row")
        return
    if not two_decimal:
        data="".join([(labels[i]+(" & {} "*len(datarow)+"\\\\\\hline\n").format(*datarow)) for i, datarow in enumerate(data)])
    else:
        if ttype in ("textCorr","textPCA","allFloat"):
            data="".join([str(labels[i])+((" & %.2f "*(len(datarow))+"\\\\\\hline\n")%tuple(datarow)) for i, datarow in enumerate(data)])
        elif ttype=="strings":
            data="".join([str(labels[i])+((" & %s "*(len(datarow))+"\\\\\\hline\n")%tuple(datarow)) for i, datarow in enumerate(data)])
        elif labels[0]=="$cc$" and len(labels)>10:
            data="".join([((labels[i]+" & %.2f "*len(datarow)+"\\\\\\hline\n")%tuple(datarow)) if labels[i] in ("$cc$","$bt$") else (((labels[i]+" & %.2f "*len(datarow)+"\\\\\n")%tuple(datarow)) if labels[i] != "$\\sigma_{dis}$" else ((labels[i]+" & %.2f "*len(datarow)+"\\\\\\hline\\hline\n")%tuple(datarow))) for i, datarow in enumerate(data)])
        elif ttype=="textGeral_":
            data="".join([str(labels[i])+((" & %.2f "*(len(datarow))+"\\\\\\hline\n")%tuple(datarow)) for i, datarow in enumerate(data)])
        elif ttype=="textCorr":
            data="".join([str(labels[i])+((" & %.2f "*(len(datarow))+"\\\\\\hline\n")%tuple(datarow)) for i, datarow in enumerate(data)])
        elif labels[0]=="$cc$" and len(labels)>5:
            data="".join([((labels[i]+" & %.2f "*len(datarow)+"\\\\\\hline\n")%tuple(datarow)) if labels[i] in ("$cc$",) 
                else 
                       (((labels[i]+" & %.2f "*len(datarow)+"\\\\\n")%tuple(datarow))           if labels[i] != "$bt$" 
                      else 
                          ((labels[i]+" & %.2f "*len(datarow)+"\\\\\\hline\\hline\n")%tuple(datarow))) for i, datarow in enumerate(data)])
        elif labels[0]=="$cc$":
            data="".join([((labels[i]+" & %.2f "*len(datarow)+"\\\\\\hline\n")%tuple(datarow)) if labels[i] in ("$asdaoijsd$",) 
                else 
                       (((labels[i]+" & %.2f "*len(datarow)+"\\\\\n")%tuple(datarow))           if labels[i] != "$bt$" 
                      else 
                          ((labels[i]+" & %.2f "*len(datarow)+"\\\\\\hline\\hline\n")%tuple(datarow))) for i, datarow in enumerate(data)])
        elif ttype=="ksDistances":
            data="".join([((str(labels[i])+" & %.3f "*4 +"\\\\\\hline\n")%tuple(datarow)) for i, datarow in enumerate(data)])
        elif ttype=="audioDistances":
            data="".join([((str(labels[i])+" & %.3f "*15 +"\\\\\\hline\n")%tuple(datarow)) for i, datarow in enumerate(data)])
        elif ttype=="textsDistances":
            data="".join([((str(labels[i])+" & %.3f "*12 +"\\\\\\hline\n")%tuple(datarow)) for i, datarow in enumerate(data)])
        elif ttype in ("audioGeneral","musicGeneral","osGeneral"):
            data="".join([((str(labels[i])+" & %s & %d " +"\\\\\\hline\n")%tuple(datarow)) for i, datarow in enumerate(data)])
        elif ttype=="textsGeneral":
            data="".join([((str(labels[i])+" & %s & %d & %d & %d & %d & %.3f & %.3f & %d & %.3f & %.3f " +"\\\\\\hline\n")%tuple(datarow)) for i, datarow in enumerate(data)])
        elif ttype=="kolmSamp":
            data="".join([((str(int(labels[i]))+" & %.3f & %.3f & %.3f & %s & %s "+" & %.3f "*6+"\\\\\\hline\n")%tuple(datarow)) for i, datarow in enumerate(data)])
        elif ttype=="kolmSamp_":
            data="".join([((str(int(labels[i]))+" & %.3f & %.3f & %.3f & %s & %s "+" & %.3f "*8+"\\\\\\hline\n")%tuple(datarow)) for i, datarow in enumerate(data)])
        elif ttype=="kolmDiff3_":
            data="".join([((str(labels[i])+" & %.3f & %.3f & %s & %s "+" & %.3f "*9+"\\\\\\hline\n")%tuple(datarow)) for i, datarow in enumerate(data)])
        elif ttype=="kolmDiff3":
            data="".join([((str(labels[i])+" & %.3f & %.3f & %.3f & %s & %s "+" & %.3f "*6+"\\\\\\hline\n")%tuple(datarow)) for i, datarow in enumerate(data)])
        elif ttype=="textGeral":
            dataFoo="".join([str(labels[:1][i])+((" & %d "*(len(datarow))+"\\\\\\hline\n")%tuple(datarow)) for i, datarow in enumerate(data[:1])])
            data=dataFoo+"".join([str(labels[1:][i])+((" & %.2f "*(len(datarow))+"\\\\\\hline\n")%tuple(datarow)) for i, datarow in enumerate(data[1:])])
        elif ttype=="textGeral__":
            data="".join([str(labels[i])+((" & %d "*(len(datarow))+"\\\\\\hline\n")%tuple(datarow)) for i, datarow in enumerate(data)])
        elif ttype=="kolmDiff2":
            data="".join([((str(labels[i])+" & %.3f "*len(datarow)+"\\\\\\hline\n")%tuple(datarow)) for i, datarow in enumerate(data)])
        elif ttype=="kolmDiff":
            #data="".join([((str(labels[i])+" & %.3f "*16+"\\\\\\hline\n")%tuple(datarow)) for i, datarow in enumerate(data)])
            data="".join([((str(labels[i])+" & %.3f & %.3f & %.3f & %s & %s "+"\\\\\\hline\n")%tuple(datarow)) for i, datarow in enumerate(data)])
        elif ttype=="kolmNull":
            try:
                data="".join([((str(labels[i])+" & %.3f & %.2f "+"& %d "*3+"\\\\\\hline\n")%tuple(datarow)) for i, datarow in enumerate(data)])
            except:
                try:
                    data="".join([((str(labels[i])+" & %.3f & %.2f "+"& %d "*4+"\\\\\\hline\n")%tuple(datarow)) for i, datarow in enumerate(data)])
                except:
                    data="".join([((str(labels[i])+" & %.3f & %.2f "+"& %d "*5+"\\\\\\hline\n")%tuple(datarow)) for i, datarow in enumerate(data)])
        elif type(data[0][0])==type("astring"):
            #data="".join([((labels[i]+" & %s "+" & %.2f "*(len(datarow)-1)+"\\\\\\hline\n")%tuple(datarow)) for i, datarow in enumerate(data)])
            data="".join([((labels[i]+" & %s "+" & %.2f "*(len(datarow)-1)+"\\\\\\hline\n")%tuple(datarow)) if type(datarow[0])==type("astring") else ((labels[i]+" & %.2f "*len(datarow)+"\\\\\\hline\n")%tuple(datarow)) for i, datarow in enumerate(data) ])
        else:
            data="".join([((str(labels[i])+" & %.2f "*len(datarow)+"\\\\\\hline\n")%tuple(datarow)) for i, datarow in enumerate(data)])
    return data

def partialSums(labels, data, partials,partial_labels="",datarow_labels=""):
    """Returns a latex table with sums of data.

    Data is though to be unidimensional. Each row
    is transposed to a column to which partial sum
    are added.
    """

    lines=[]
    for label in labels:
        lines.append("{} ".format(label))
    for datarow in data:
        for partial in partials:
            for line_num in range(len(datarow)):
                if (line_num%partial)==0:
                    lines[line_num]+="& \\multirow{{{:d}}}{{*}}{{ {:.2f} }}  ".format(partial,sum(datarow[line_num:line_num+partial]))
                else:
                    lines[line_num]+="& "

    for line_num in range(len(lines)):
        cuts=[(0==(line_num+1)%partial) for partial in partials]
        i=0
        suffix=""
        for cut in cuts:
            if cut:
                for datarownum in range(len(data)):
                    num=i+(datarownum)*len(partials)
                    suffix+="\\cline{{{}-{}}}".format(num+2,num+2)
            i+=1
        lines[line_num]+="\\\\{}\n".format(suffix)
    
    ltable="".join(lines)

    if partial_labels:
        header=( (" & {}"*len(partial_labels)).format(*partial_labels) )*len(data)+" \\\\\\hline\n"
        ltable=header+ltable
    if datarow_labels:
        header=((" & \\multicolumn{{%i}}{{c|}}{{{}}}"%(len(partials),))*len(datarow_labels)).format(*datarow_labels)+" \\\\\\hline\n"
        ltable=header+ltable
    header="\\begin{center}\n\\begin{tabular}{| l ||"+" c |"*len(data)*len(partials)+"}\\hline\n"
    footer="\\hline\\end{tabular}\n\\end{center}"
    ltable=header+ltable+footer
    return ltable

def pcaTable(labels,vec_mean,vec_std,val_mean,val_std):
    """Make table with PCA formation mean and std"""

    header="\\begin{center}\n\\begin{tabular}{| l |"+" c |"*6+"}\\cline{2-7}\n"
    header+="\\multicolumn{1}{c|}{} & \\multicolumn{2}{c|}{PC1}          & \multicolumn{2}{c|}{PC2} & \multicolumn{2}{c|}{PC3}  \\\\\\cline{2-7}"
    header+="\\multicolumn{1}{c|}{} & $\mu$            & $\sigma$ & $\mu$         & $\sigma$ & $\mu$ & $\sigma$  \\\\\\hline\n"
    tt=n.zeros((vec_mean.shape[0],6))
    tt[:,::2]=vec_mean
    tt[:,1::2]=vec_std
    tt_=n.zeros(6)
    tt_[::2]=val_mean
    tt_[1::2]=val_std
    tab_data=n.vstack((tt,tt_))
    footer="\\hline\\end{tabular}\n\\end{center}"
    table=header + makeTables(labels,tab_data,True) + footer
    return table
def vstackTables_(fname1,fname2,fname3):
    s1=open(fname1+".tex").read()
    s2=open(fname2+".tex").read()
    s1_=re.split("\n\\\\end{tabular",s1)[0]
    s2_=re.split("tabular.*\\hline\n &.*g.*hline",s2)[1]
    writeTex(s1_+"\\hline\\hline"+s2_,fname3+".tex")

def vstackTables(fname1,fname2,fname3):
    s1=open(fname1+".tex").read()
    s2=open(fname2+".tex").read()
    s1_=re.split("\n\\\\end{tabular",s1)[0]
    s2_=re.split("tabular.*\\hline\n &.*g.*hline",s2)[1]
    writeTex(s1_+"\\hline"+s2_,fname3+".tex")
def writeTex(string,filename):
    with open(filename,"w") as f:
        f.write(string)
def dl(fname,hl,vl,hl_=[],over=1):
    fname+=".tex"
    foo=doubleLines(fname,hlines=hl,vlines=vl,hlines_=hl_)
    if not over:
        fname=fname.replace(".tex","_.tex")
    writeTex(foo,fname)
def me(fname,mark,locs,over=0):
    fn=fname+".tex"
    foo=markEntries_(fn,mark,locs)
    if not over:
        fn=fn.replace(".tex","_.tex")
    writeTex(foo,fn)





import sys
import collections as c
from scipy import special, stats
import numpy as n, pylab as p, networkx as x

class NetworkDrawer:
    drawer_count=0
    def __init__(self,metric="strength"):
        self.drawer_count+=1
        metric_=self.standardizeName(metric)
        self.metric_=metric_
        self.draw_count=0

    def standardizeName(self,name):
        if name in (["s","strength","st"]+["f","força","forca","fo"]):
            name_="s"
        if name in (["d","degree","dg"]+["g","grau","gr"]):
            name_="d"
        return name_
    def makeLayout(self,network_measures,network_partitioning=None):
        """Delivers a sequence of user_ids and (x,y) pos.
        """
        self.network_measures=network_measures
        if self.metric_=="s":
            measures_=network_measures.strengths
        elif self.metric_=="d":
            measures_=network_measures.degrees
        else:
            print("not known metric to make layout")
        self.ordered_measures=ordered_measures = c.OrderedDict(sorted(measures_.items(), key=lambda x: x[1]))
        self.measures=measures=list(ordered_measures.values())
        self.authors=authors=  list(ordered_measures.keys())
        total=network_measures.N
        if not network_partitioning:
            self.k1=k1=round(total*.80)
            self.k2=k2=round(total*.95)
            self.periphery=authors[:k1]
            self.intermediary=authors[k1:k2]
            self.hubs=authors[k2:]
        else:
            sectors=network_partitioning.sectorialized_agents__
            self.k1=k1=len(sectors[0])
            self.k2=k2=k1+len(sectors[1])
            self.periphery,self.intermediary,self.hubs=sectors
        print("fractions ={:0.4f}, {:0.4f}, {:0.4f}".format(k1/total, (k2-k1)/total, 1-k2/total))
        self.makeXY()

    def drawNetwork(self, network,network_measures,filename="example.png",label="auto",network_partitioning=None):
        p.clf()

        if self.metric_=="s":
            measures_=network_measures.strengths
        elif self.metric_=="d":
            measures_=network_measures.degree
        else:
            print("not known metric to make layout")
        ordered_measures = c.OrderedDict(sorted(measures_.items(), key=lambda x: x[1]))
        measures=list(ordered_measures.values())
        authors=  list(ordered_measures.keys())

        total=network_measures.N
        if not network_partitioning:
            k1=k1=round(total*.80)
            k2=k2=round(total*.95)
            periphery=authors[:k1]
            intermediary=authors[k1:k2]
            hubs=authors[k2:]
        else:
            sectors=network_partitioning.sectorialized_agents__
            k1=k1=len(sectors[0])
            k2=k2=k1+len(sectors[1])
            periphery,intermediary,hubs=(set(iii) for iii in sectors)

        in_measures=network_measures.in_strengths
        min_in=max(in_measures.values())/3+0.1
        out_measures=network_measures.out_strengths
        min_out=max(out_measures.values())/3+.1

        self.clustering=clustering=network_measures.weighted_clusterings
        A=x.drawing.nx_agraph.to_agraph(network.g)
        A.node_attr['style']='filled'
        A.graph_attr["bgcolor"]="black"
        A.graph_attr["pad"]=.1
        #A.graph_attr["size"]="9.5,12"
        A.graph_attr["fontsize"]="25"
        if label=="auto":
            label=self.makeLabel()
        A.graph_attr["label"]=label
        A.graph_attr["fontcolor"]="white"
        cm=p.cm.Reds(range(2**10)) # color table
        self.cm=cm
        nodes=A.nodes()
        self.colors=colors=[]
        self.inds=inds=[]
        self.poss=poss=[]
        for node in nodes:
            n_=A.get_node(node)
            ind_author=self.authors.index(n_)
            inds.append(inds)
            colors.append(        '#%02x%02x%02x' % tuple([int(255*i) for i in cm[int(clustering[n_]*255)][:-1]]))
            #n_.attr['fillcolor']= '#%02x%02x%02x' % tuple([255*i for i in cm[int(clustering[n_]*255)][:-1]])
            n_.attr['fillcolor']= colors[-1]
            n_.attr['fixedsize']=True
            n_.attr['width']=  abs(.6*(in_measures[n_]/min_in+  .05))
            n_.attr['height']= abs(.6*(out_measures[n_]/min_out+.05))
            if n_ in hubs:
                n_.attr["shape"] = "hexagon"
            elif n_ in intermediary:
                pass
            else:
                n_.attr["shape"] = "diamond"
            pos="%f,%f"%tuple(self.posXY[ind_author])
            poss.append(pos)
            n_.attr["pos"]=pos
            n_.attr["pin"]=True
            n_.attr["fontsize"]=25
            n_.attr["fontcolor"]="white"
            n_.attr["label"]=""

        weights=[s[2]["weight"] for s in network_measures.edges]
        self.weights=weights
        max_weight=max(weights)
        self.max_weight=max_weight
        self.weights_=[]
        edges=A.edges()
        for e in edges:
            factor=float(e.attr['weight'])
            self.weights_.append(factor)
            e.attr['penwidth']=.34*factor
            e.attr["arrowsize"]=1.5
            e.attr["arrowhead"]="lteeoldiamond"
            w=factor/max_weight # factor em [0-1]

            cor=p.cm.Spectral(int(w*255))
            self.cor=cor
            cor256=255*n.array(cor[:-1])

            r0=int(cor256[0]/16)
            r1=int(cor256[0]-r0*16)
            r=hex(r0)[-1]+hex(r1)[-1]

            g0=int(cor256[1]/16)
            g1=int(cor256[1]-g0*16)
            g=hex(g0)[-1]+hex(g1)[-1]

            b0=int(cor256[2]/16)
            b1=int(cor256[2]-b0*16)
            b=hex(b0)[-1]+hex(b1)[-1]

            #corRGB="#"+r+g+b+":#"+r+g+b
            corRGB="#"+r+g+b

            e.attr["color"]=corRGB

        A.draw(filename, prog="neato") # twopi ou circo
        ################
        self.A=A
        self.draw_count+=1
    def makeLabel(self):
        label=""
        if "window_size" in dir(self):
            label+="w: {}, ".format(self.window_size)
            #m: %i, N = %i, E = %i"%(self.draw_count*self.step_size,self.network_measures.N,self.network_measures.E)
        if "step_size" in dir(self):
            label+="m: {} ,".format(self.draw_count*self.step_size+self.offset)
        else:
            label+="m: %i, ".format(self.draw_count)
            #self.network_measures.N,self.network_measures.E)
        label+="N = %i, E = %i"%(self.network_measures.N,self.network_measures.E)
        return label

    def updateNetwork(self,network,networkMeasures=None):
        pass
    def makeXY(self):
        size_periphery=self.k1
        size_intermediary=self.k2-self.k1
        size_hubs=self.network_measures.N-self.k2
        if size_hubs%2==1:
            size_hubs+=1
            size_intermediary-=1
        xh=n.linspace(0,0.5,size_hubs,endpoint=False)[::-1]
        thetah=2*n.pi*xh
        yh=n.sin(thetah)

        xi=n.linspace(1,0.5, size_intermediary, endpoint=True)
        thetai=2*n.pi*xi
        yi=n.sin(thetai)

        xp=n.linspace(.95,0.4, size_periphery)[::-1]
        yp=n.linspace(.1,1.25, size_periphery)[::-1]

        self.pos=((xp,yp),(xi,yi),(xh,yh))
        XFACT=7
        YFACT=3
        self.posX=posX=n.hstack((xp,xi,xh))*XFACT
        self.posY=posY=n.hstack((yp,yi,yh))*YFACT
        self.posXY=n.vstack((posX.T,posY.T)).T



import collections as c, numpy as n
class AgentStatistics:
    def __init__(self,list_datastructures=None):
        if not list_datastructures:
            print("input datastructures, please")
        self.list_datastructures=list_datastructures

        self.author_messages_= author_messages_= c.OrderedDict(sorted(list_datastructures.author_messages.items(), key=lambda x: len(x[1])))
        self.n_authors=len(author_messages_)

        self.authors=list(author_messages_.keys())
        self.msgs=list(author_messages_.values())

        self.authors_n_msgs=n.array([len(i) for i in self.msgs])

        self.basicMeasures()
        self.messageAuthorsCorrespondence()
    def basicMeasures(self):
        self.mean_msgs=n.mean(self.authors_n_msgs)
        self.std_msgs=n.std(self.authors_n_msgs)
        plist=[5,10,25,50,75,90,95]
        self.percentiles=percentiles={}
        for percentile in plist:
            percentiles[percentile]=n.percentile(self.authors_n_msgs,plist)

    def oneMore(self,bool_array):
        args=n.nonzero(bool_array-1)[0]
        bool_array[args[0]]=True
        return bool_array
    def oneLess(self,bool_array):
        args=n.nonzero(bool_array-1)[0]
        bool_array[args[-1]]=True
        return bool_array
    def messageAuthorsCorrespondence(self):
        self.n_msgs_h=self.authors_n_msgs[-1]
        self.n_msgs_h_=100*self.n_msgs_h/self.list_datastructures.n_messages
        self.cumulative=n.array([n.sum(self.authors_n_msgs[:i+1]) for i in range(self.n_authors)])
        self.cumulative_=self.cumulative/self.list_datastructures.n_messages
        self.last_d10=1+(self.cumulative_<.10).sum()
        self.last_d10_=self.last_d10/self.n_authors
        self.q1=1+(self.cumulative_>0.75).sum()
        self.q1_=self.q1/self.n_authors
        self.q3=1+(self.cumulative_>0.25).sum()
        self.q3_=self.q3/self.n_authors

        self.Mlast_d10 =self.authors_n_msgs[self.oneMore(self.cumulative_<0.10)].sum()/self.list_datastructures.n_messages

        self.Mq1=self.authors_n_msgs[self.oneLess(self.cumulative_>0.75)].sum()/self.list_datastructures.n_messages

        self.Mq3=self.authors_n_msgs[self.oneLess(self.cumulative_>0.25)].sum()/self.list_datastructures.n_messages




import numpy as n, pylab as p, matplotlib
from scipy import stats

class NetworkPCA:
    """PCA cases of interest  for observing stability in interaction networks.

        1) Primacy of centrality measures for dispersion; and
        2) precedence of symmetry over clustering;
        3) average of all basic centrality measures for first compoment
    """
    def __init__(self,network_measures,network_partitioning=None,tdir=".",tname="sample.png", measures="all",plot_sym=False):
        # enable selection of measures input as string
        # through measures variable and exec() or eval() methods.
        self.M1=n.array(( network_measures.weighted_clusterings_,
                    network_measures.degrees_,
                    network_measures.weighted_directed_betweenness_
                    ))
        self.pca1=PCA(self.M1)
        if "in_strengths_" in dir(network_measures):
            self.M2=n.array(( network_measures.weighted_clusterings_,
                        network_measures.strengths_,
                        network_measures.in_strengths_,
                        network_measures.out_strengths_,
                        network_measures.degrees_,
                        network_measures.in_degrees_,
                        network_measures.out_degrees_,
                        network_measures.weighted_directed_betweenness_
                        ))
            self.M3=n.array(( network_measures.weighted_clusterings_,
                        network_measures.strengths_,
                        network_measures.in_strengths_,
                        network_measures.out_strengths_,
                        network_measures.degrees_,
                        network_measures.in_degrees_,
                        network_measures.out_degrees_,
                        network_measures.weighted_directed_betweenness_,
                        network_measures.asymmetries,
                        network_measures.asymmetries_edge_mean,
                        network_measures.asymmetries_edge_std,
                        network_measures.disequilibrium,
                        network_measures.disequilibrium_edge_mean,
                        network_measures.disequilibrium_edge_std,
                        ))
            self.pca2=PCA(self.M2)
            self.pca3=PCA(self.M3)
        if plot_sym:
            self.pca3.plotSym(network_partitioning,tdir,tname)
#            fig = matplotlib.pyplot.gcf()
#            fig.set_size_inches(11.,8.4)
#            p.suptitle("Symmetry prevalence over clutering for data dispersion")
#            p.subplot(311)
#            # plot degree x clust
#            p.title("degree x clustering")
#            label1=r"degree $\rightarrow$"
#            label2=r"clustering $\rightarrow$"
#            p.xlabel(label1, fontsize=15)
#            p.ylabel(label2, fontsize=15)
#            n_periphery=   len(network_partitioning.sectorialized_agents__[0])
#            n_intermediary=len(network_partitioning.sectorialized_agents__[1])
#            n_hubs=        len(network_partitioning.sectorialized_agents__[2])
#            M=n.array(network_measures.degrees_)
#            network_measures.degrees__=(M-M.mean())/M.std()
#            M=n.array(network_measures.weighted_clusterings_)
#            network_measures.weighted_clusterings__=(M-M.mean())/M.std()
#            p.ylim(min(network_measures.weighted_clusterings__)-1,max(network_measures.weighted_clusterings__)+1)
#            p.xlim(min(network_measures.degrees__)-1,max(network_measures.degrees__)+1)
#            p.plot(network_measures.degrees__[:n_periphery],network_measures.weighted_clusterings__[:n_periphery],"ko", ms=3.9,label="periphery")
#
#            p.plot(network_measures.degrees__[n_periphery:n_periphery+n_intermediary],network_measures.weighted_clusterings__[n_periphery:n_periphery+n_intermediary],"k^", ms=3.9,label="intermediary")
#
#            p.plot(network_measures.degrees__[-n_hubs:],network_measures.weighted_clusterings__[-n_hubs:],"k*", ms=3.9,label="hubs")
#
#            #p.subplot(312)
#            #self.pca2.plot(None,network_partitioning,labels=None,tdir=None,savefig=False,clear_fig=False,title="Vertices in principal components",label1=r"PC1 - degrees and strengths",label2=r"PC3 - clustering")
#            p.subplot(212)
#            self.pca3.plot(None,network_partitioning,labels=None,tdir=None,savefig=False,clear_fig=False,title="Vertices in principal components",label1=r"PC1 - degrees and strengths",label2="PC2 - symmetry")
#            p.subplots_adjust(left=0.08,bottom=0.12,right=0.97,top=0.88,wspace=0.13,hspace=0.88)
#            p.show()
#            #p.savefig("{}/{}".format(tdir,tname))
        elif network_partitioning:
            self.pca1.plot(tname.replace(".","_1."),network_partitioning,tdir=tdir)
            self.pca2.plot(tname.replace(".","_2."),network_partitioning,tdir=tdir)
            self.pca3.plot(tname.replace(".","_3."),network_partitioning,tdir=tdir)


class PCA:
    """Apply PCA to incoming datatable M (metrics x observations)

    Usage
    =====

    Initialize with a n x m matrix of n metrics each with m observations
    >>> foo=n.random(100)
    >>> p1=n.vstack((foo,foo))
    >>> p2=n.vstack((-foo,foo))
    >>> p3=n.vstack((foo,-foo))
    >>> M=n.hstack((p1,p2,p3))
    >>> pca=g.PCA(M)

    See attributes for information about data:
    >>> pca.eig_values # for eigen values from greatest down
    >>> pca.eig_values_ # for a normalized eig_values
    >>> pca.eig_vectors # for eigen vectors of the eig_values
    >>> pca.eig_vectors_ # for a normalized eig_vectors
    >>> pca.C # for covariance matrix
    >>> pca.M # for initial data 
    >>> pca.x # final positions in the principal component
    >>> pca.y # final positions in second principal component
    >>> pca.plot() # to plot observations in initial and final spaces
    

    """
    def __init__(self,*metrics,final_dimensions=2,draw=False):
        M=n.vstack(metrics)
        # zscore: # USE NUMPY.stats.zscore(M, axis=1, ddof=1)
        self.M_=n.copy(M)
        for i in range(M.shape[0]):
            if M[i].std():
                M[i]=(M[i]-M[i].mean())/M[i].std()
            else:
                M[i]=0.
        # convariance matrix:
        self.C=n.cov(M,bias=1)
        self.M=M

        eig_values, eig_vectors = n.linalg.eig(self.C)
        # Ordering eigenvalues and eigenvectors
        args=n.argsort(eig_values)[::-1]
        self.eig_values=eig_values[args]
        self.eig_values_=100*self.eig_values/n.sum(n.abs(self.eig_values))
        self.eig_vectors=eig_vectors[:,args]
        self.eig_vectors_=n.array([100*self.eig_vectors[:,i]/n.abs(self.eig_vectors[:,i]).sum() for i in range(self.eig_vectors.shape[1])]).T
        # retaining only some eigenvectors
        self.feature_vec=self.eig_vectors[:,:final_dimensions]
        self.feature_vec_=n.array([100*self.feature_vec[:,i]/n.abs(self.feature_vec[:,i]).sum() for i in range(self.feature_vec.shape[1])]).T

        self.final_data=n.dot(M.T,self.feature_vec)
        self.x=self.final_data[:,0]
        self.y=self.final_data[:,1]
    #def plot(self, tname="sample.png", network_partitioning=False,labels="full", tdir=".",savefig=True,clear_fig=True,title="Vertex plot in principal components (PCA)",label1="PC1",label2="PC2"):
    def plotSym(self,network_partitioning,tdir,tname):

        self.x_=self.M_[4] # degrees
        self.cc=self.M[0] #clustering 
        p.figure(figsize=(7.,4.))
        p.subplots_adjust(left=0.09,bottom=0.16,right=0.95,top=0.87,hspace=0.04)
        #fig = matplotlib.pyplot.gcf()
        #fig.set_size_inches(6.,8.4)
        #p.suptitle("Symmetry-related measures and clutering coefficient along connectivity")
        kl=max(network_partitioning.sectorialized_degrees__[0])+.5
        kr=max(network_partitioning.sectorialized_degrees__[1])+.5
        p.suptitle("Symmetry-related and clutering coefficient components along connectivity")
        p.subplot(211)
        p.plot((kl,kl),(-1000,1000),"g--",linewidth=3)
        p.plot((kr,kr),(-1000,1000),"g--",linewidth=3)
        p.xticks((),())
            # plot degree x clust
        #label1=r"degree $\rightarrow$"
        label2=r"clustering $\rightarrow$"
        #p.xlabel(label1, fontsize=15)
        p.ylabel(label2, fontsize=15)
        n_periphery=   len(network_partitioning.sectorialized_agents__[0])
        n_intermediary=len(network_partitioning.sectorialized_agents__[1])
        n_hubs=        len(network_partitioning.sectorialized_agents__[2])
        #M=n.array(network_measures.degrees_)
        #network_measures.degrees__=(M-M.mean())/M.std() #AQUI
        #M=n.array(network_measures.weighted_clusterings_)
        #network_measures.weighted_clusterings__=(M-M.mean())/M.std()
        p.ylim(min(self.cc)-0.9,max(self.cc)+.9)
        p.xlim(min(self.x_)-0.1,max(self.x_)+.1)
        p.plot(self.x_[:n_periphery],                          self.cc[:n_periphery],"ko", label="periphery"                             ,ms=10,alpha=.4)
        p.plot(self.x_[n_periphery:n_periphery+n_intermediary],self.cc[n_periphery:n_periphery+n_intermediary],"k^", label="intermediary",ms=10,alpha=.4)
        p.plot(self.x_[n_periphery+n_intermediary:],                              self.cc[-n_hubs:],"k*", label="hubs"                                      ,ms=10,alpha=.4)
        #p.plot(self.x_[-n_hubs:],                              self.cc[-n_hubs:],"k*", label="hubs"                                      ,ms=10,alpha=.4)

        #p.legend()
        p.legend(bbox_to_anchor=(0.17, .71, .8, .2), loc=3,
                           ncol=3, mode="expand", borderaxespad=0.)
        p.subplot(212)
        p.plot((kl,kl),(-1000,1000),"g--",linewidth=3)
        p.plot((kr,kr),(-1000,1000),"g--",linewidth=3)
        p.xticks((kl,kr),(r"k_L",r"k_R"))
        p.ylabel(r"PC2 $\rightarrow$", fontsize=15)
        p.xlabel(r"degree $\rightarrow$", fontsize=15)
#        p.title("degree x symmetry-related principal component")
        #self.pca3.plot(None,network_partitioning,labels=None,tdir=None,savefig=False,clear_fig=False,title="Vertices in principal components",label1=r"PC1 - degrees and strengths",label2="PC2 - symmetry")
        #p.subplots_adjust(left=0.08,bottom=0.12,right=0.97,top=0.88,wspace=0.13,hspace=0.88)
        #p.show()
        #p.savefig("{}/{}".format(tdir,tname))
#        elifnetwork_partitioning:
#            self.pca1.plot(tname.replace(".","_1."),network_partitioning,tdir=tdir)
#            self.pca2.plot(tname.replace(".","_2."),network_partitioning,tdir=tdir)
#            self.pca3.plot(tname.replace(".","_3."),network_partitioning,tdir=tdir)
#        p.title(title)

        p.ylim(min(self.y) -0.9,max(self.y)+  0.9)
        p.xlim(min(self.x_)-0.1,max(self.x_)+0.1)
        #n_periphery=   len(network_partitioning.sectorialized_agents__[0])
        #n_intermediary=len(network_partitioning.sectorialized_agents__[1])
        #n_hubs=        len(network_partitioning.sectorialized_agents__[2])
        p.plot(self.x_[:n_periphery],self.y[:n_periphery],"ko", label="perihpery"                                                       ,ms=10,alpha=.4)
        p.plot(self.x_[n_periphery:n_periphery+n_intermediary],self.y[n_periphery:n_periphery+n_intermediary],"k^", label="intermediary",ms=10,alpha=.4)
        p.plot(self.x_[-n_hubs:],self.y[-n_hubs:],"k*", label="hubs"                                                                    ,ms=10,alpha=.4)
        #p.show()
        p.savefig("{}/{}".format(tdir,tname))

    def plot(self, tname="sample.png", network_partitioning=False,labels="full", tdir=".",savefig=True,clear_fig=True,title="Vertex plot in principal components (PCA)",label1="PC1",label2="PC2"):
        if clear_fig:
            p.clf()
        if labels=="full":
            #foo=self.feature_vec[:,0]
            #foo_=("%.2f, "*len(foo)) % tuple(foo)
            foo=self.feature_vec_[:,0]
            foo__=("%.2f, "*len(foo)) % tuple(foo)
            label1+=" " + foo__

            #foo=self.feature_vec[:,1]
            #foo_=("%.2f, "*len(foo)) % tuple(foo)
            foo=self.feature_vec_[:,1]
            foo__=("%.2f, "*len(foo)) % tuple(foo)
            label2+=" " +foo__

            #foo=(self.eig_values[:4]/self.eig_values.sum())*100
            #foo_=r"$\lambda = $"+("%.2f, "*len(foo) % tuple(foo))
            foo=(self.eig_values_[:4])
            foo__=r"$\lambda = $"+("%.2f, "*len(foo) % tuple(foo))
            title+=" "+foo__
        if labels=="sym":
            pass

        p.xlabel(label1, fontsize=15)
        p.ylabel(label2, fontsize=15)
        #p.title(foo_)
        p.title(title)

        p.ylim(min(self.y)-1,max(self.y)+1)
        p.xlim(min(self.x)-1,max(self.x)+1)
        if not network_partitioning:
            p.plot(self.x,self.y,"k^", ms=3.9,label="intermediary")
        else:
            n_periphery=   len(network_partitioning.sectorialized_agents__[0])
            n_intermediary=len(network_partitioning.sectorialized_agents__[1])
            n_hubs=        len(network_partitioning.sectorialized_agents__[2])
            p.plot(self.x[:n_periphery],self.y[:n_periphery],"k^", ms=3.9,label="perihpery")
            p.plot(self.x[n_periphery:n_periphery+n_intermediary],self.y[n_periphery:n_periphery+n_intermediary],"k*", ms=3.9,label="hubs")
            p.plot(self.x[-n_hubs:],self.y[-n_hubs:],"ro", ms=3.9,label="hubs")
        p.legend()
        if savefig:
            p.savefig("{}/{}".format(tdir,tname))
            x=self.M[0]
            y=self.M[1]
            if clear_fig:
                p.clf()
            p.plot(x,y,"go", ms=3.9,label="intermediary")
            p.savefig("{}/Initial{}.png".format(tdir,tname))



import numpy as n, calendar, datetime

def circularStatistics(population, period):
    pop=n.array(population)
    pop_=pop*2*n.pi/period
    j=n.complex(0,1)
    vec_n=n.e**(j*pop_)
    mean_vec=vec_n.sum()/len(vec_n) # first moment
    mean_angle=n.arctan2(mean_vec.real,mean_vec.imag)
    size_mean_vec=n.abs(mean_vec)
    variance_unity_radius=1-size_mean_vec
    std_unity_radius=n.sqrt(-2*n.log(size_mean_vec))
    circular_mean=mean_angle*(period/(2*n.pi))
    circular_variance=variance_unity_radius*(period**2/(2*n.pi))
    circular_std=std_unity_radius*(period/(2*n.pi))

    second_moment=(vec_n**2).sum()/len(vec_n)
    size_second_moment=n.abs(second_moment)
    circular_dispersion=(1-size_second_moment)/(2*(size_mean_vec**2))

    return dict(mean_vec=mean_vec,
            mean_angle=mean_angle, 
            size_mean_vec=size_mean_vec,
            circular_mean=circular_mean, 
            circular_variance=circular_variance, 
            circular_std=circular_std, 
            variance_unity_radius=variance_unity_radius, 
            std_unity_radius=std_unity_radius,
            circular_dispersion=circular_dispersion)

class TimeStatistics:
    def __init__(self,list_datastructures=None):
        if not list_datastructures:
            print("input datastructures, please")
        datetimes=[]
        for datetime in list_datastructures.raw_clean_dates:
            datetimes.append(datetime[1])
        self.datetimes=datetimes
        self.n_observations=len(datetimes)
        self.bad_datetimes=[]
        self.makeStatistics()
    def makeStatistics(self):
        """Make statistics from seconds to years"""
        self.uniformComparisson()
        self.secondsStats()
        self.minutesStats()
        self.hoursStats()
        self.weekdaysStats()
        self.monthdaysStats_()
        self.monthsStats()
        self.yearsStats()
    def uniformComparisson(self):
        ar=n.random.randint(0,60,(1000,self.n_observations))
        cc=n.array([n.histogram(i,60)[0] for i in ar])
        cc_=cc.min(1)/cc.max(1)
        self.obs60=(cc_.mean(),cc_.std())
        self.obs60_=cc_

        ar=n.random.randint(0,24,(1000,self.n_observations))
        cc=n.array([n.histogram(i,24)[0] for i in ar])
        cc_=cc.min(1)/cc.max(1)
        self.obs24=(cc_.mean(),cc_.std())
        self.obs24_=cc_

        ar=n.random.randint(0,30,(1000,self.n_observations))
        cc=n.array([n.histogram(i,30)[0] for i in ar])
        cc_=cc.min(1)/cc.max(1)
        self.obs30=(cc_.mean(),cc_.std())
        self.obs30_=cc_

        ar=n.random.randint(0,7,(1000,self.n_observations))
        cc=n.array([n.histogram(i,7)[0] for i in ar])
        cc_=cc.min(1)/cc.max(1)
        self.obs7=(cc_.mean(),cc_.std())
        self.obs7_=cc_
        
        #self.obs60=n.random.randint(0,60,(1000, self.n_observations))
        #self.count_obs60=[obs60.count(i) for i in set(obs60)]
        #self.obs24=n.random.randint(0,24,self.n_observations)
        #self.count_obs24=[obs24.count(i) for i in set(obs24)]
        # IN MONTHs function:
        #self.obs12=n.random.randint(0,12,len(self.months.samples))
        #self.obs30=n.random.randint(0,30,self.n_observations)
        #self.count_obs12=[obs12.count(i) for i in set(obs12)]
        #self.obs7=n.random.randint(0,7,self.n_observations)
        #self.count_obs12=[obs12.count(i) for i in set(obs12)]
        
    def secondsStats(self):
        # contagem para histograma
        seconds=[i.second for i in self.datetimes]
        histogram=n.histogram(seconds,bins=list(range(61)))[0]
        max_discrepancy=histogram.min()/histogram.max()
        # medidas circulares
        circular_measures=circularStatistics(seconds,60)
        seconds=dict(
            circular_measures=circular_measures,
            max_discrepancy=max_discrepancy,
            max_discrepancy_=self.obs60,
            samples=seconds,
            histogram=histogram)
        self.seconds=seconds

    def minutesStats(self):
        samples=[i.minute for i in self.datetimes]
        histogram=n.histogram(samples,bins=list(range(61)))[0]
        max_discrepancy=histogram.min()/histogram.max()
        # medidas circulares
        circular_measures=circularStatistics(samples,60)
        minutes=dict(
            samples=samples,
            histogram=histogram,
            max_discrepancy=max_discrepancy,
            max_discrepancy_=self.obs60,
            circular_measures=circular_measures
        )
        self.minutes=minutes


    def hoursStats(self):
        samples=[i.hour for i in self.datetimes]
        histogram=n.histogram(samples,bins=list(range(25)))[0]
        max_discrepancy=histogram.min()/histogram.max()
        # medidas circulares
        circular_measures=circularStatistics(samples,24)
        hours=dict(
            samples=samples,
            histogram=histogram,
            max_discrepancy=max_discrepancy,
            max_discrepancy_=self.obs24,
            circular_measures=circular_measures
        )
        self.hours=hours
    def weekdaysStats(self):
        samples=[i.weekday() for i in self.datetimes]
        histogram=n.histogram(samples,bins=list(range(8)))[0]
        max_discrepancy=histogram.min()/histogram.max()
        # medidas circulares
        circular_measures=circularStatistics(samples,7)
        self.weekdays=dict(
            samples=samples,
            histogram=histogram,
            max_discrepancy=max_discrepancy,
            max_discrepancy_=self.obs7,
            circular_measures=circular_measures
        )
    def monthdaysStats_(self):
        def aux(xx):
            #return (xx.day-1)/(
            #        calendar.monthrange(xx.year, xx.month)[1] )
            return ((xx.day-1)*24*60+xx.hour*60+xx.minute )/(
                    calendar.monthrange(xx.year, xx.month)[1]*24*60)
        samples=[aux(i) for i in self.datetimes]
        mean_month_size=n.mean([calendar.monthrange(xx.year, xx.month)[1]
            for xx in self.datetimes])
        mean_month_size=n.round(mean_month_size)
        histogram=n.histogram(samples,bins=n.linspace(0,1,mean_month_size+1))[0]
        max_discrepancy=histogram.min()/histogram.max()
        # medidas circulares
        circular_measures=circularStatistics([i*mean_month_size for i in samples],mean_month_size)
        self.monthdays=dict(
            mean_month_size=mean_month_size,
            samples=samples,
            histogram=histogram,
            max_discrepancy=max_discrepancy,
            max_discrepancy_=self.obs30,
            circular_measures=circular_measures,
        )
    def monthdaysStats(self):
        def aux(xx):
            return (xx.day-1)/(
                    calendar.monthrange(xx.year, xx.month)[1] )
        samples=[aux(i) for i in self.datetimes]
        mean_month_size=n.mean([calendar.monthrange(xx.year, xx.month)[1]
            for xx in self.datetimes])
        mean_month_size=n.round(mean_month_size)
        histogram=n.histogram(samples,bins=n.linspace(0,1,mean_month_size+1))[0]
        max_discrepancy=histogram.min()/histogram.max()
        # medidas circulares
        circular_measures=circularStatistics(samples,1)
        self.monthdays=dict(
            mean_month_size=mean_month_size,
            samples=samples,
            histogram=histogram,
            max_discrepancy=max_discrepancy,
            max_discrepancy_=self.obs30,
            circular_measures=circular_measures,
        )
    def monthsStats(self,truncate=True):
        year=365.242199 # days
        if truncate:
            delta=self.datetimes[-1]-self.datetimes[0]
            if delta.days > year:
                delta_=(delta.total_seconds()/(24*60*60))%year
                max_date=self.datetimes[-1]-datetime.timedelta(delta_%year)
            else:
                max_date=self.datetimes[-1]
            try:
                samples=[i.month-1 for i in self.datetimes if i <= max_date]
            except:
                samples=[]
                for i in self.datetimes:
                    try:
                        foo=i<=max_date
                        if foo:
                            samples.append(i.month-1)
                    except:
                        self.bad_datetimes+=[i]

        else:
            samples=[i.month-1 for i in self.datetimes]
        histogram=n.histogram(samples,bins=list(range(13)))[0]
        max_discrepancy=histogram.min()/histogram.max()
        # medidas circulares
        circular_measures=circularStatistics(samples,12)

        ar=n.random.randint(0,12,(1000,len(samples)))
        cc=n.array([n.histogram(i,12)[0] for i in ar])
        cc_=cc.min(1)/cc.max(1)
        self.obs12=(cc_.mean(),cc_.std())
        self.obs12_=cc_


        self.months=dict(
            samples=samples,
            histogram=histogram,
            max_discrepancy=max_discrepancy,
            max_discrepancy_=self.obs12,
            circular_measures=circular_measures
        )
    def yearsStats(self):
        samples=[i.year for i in self.datetimes]
        smin=min(samples)
        smax=max(samples)
        histogram=n.histogram(samples,bins=list(range(smin,smax+2)))[0]
        max_discrepancy=histogram.min()/histogram.max()
        self.years=dict(
            samples=samples,
            histogram=histogram,
            max_discrepancy=max_discrepancy,
        )



import networkx as x, time as t, numpy as n
class NetworkMeasures:
    """Extracts measures for incoming network.

    Both overall measures and measures of each vertex are taken.
    """
    def __init__(self,network=None,exclude=["unweghted_undirected_betweenness","weighted_undirected_betweenness","unweighted_directed_betweenness", "rich_club"]):
        if not network:
            network=g.InteractionNetwork()

        #_all=["clustering"] # measures taken from graph by a method of itself
        #_all_=[] # measures taken from graph by a networkx method
        #_allu=["clustering"] # measures taken from an undirected version of the graph by a method of itself
        #_allu_=[] # measures taken from the undirected version of the graph by a networkx method
        self.makeMeasures(network, exclude)
    def makeMeasures(self,network,exclude):
        """Make the network measures"""
        # fazer condicional para cada medida, se não estiver na exclude[],
        # fazer medida de tempo e guardar como tupla no

        g=network.g
        gu=network.gu
        timings=[]

        T=t.time()
        self.N=network.g.number_of_nodes()
        self.E=network.g.number_of_edges()
        self.E_=network.gu.number_of_edges()
        self.edges=     g.edges(data=True)
        self.nodes=     g.nodes(data=True)
        timings.append((t.time()-T,"edges and nodes"))

        T=t.time()
        self.degrees=dict(g.degree())
        self.nodes_= sorted(g.nodes(), key=lambda x : self.degrees[x])
        self.degrees_=[self.degrees[i] for i in self.nodes_]
        self.in_degrees=dict(g.in_degree())
        self.in_degrees_=[self.in_degrees[i] for i in self.nodes_]
        self.out_degrees=dict(g.out_degree())
        self.out_degrees_=[self.out_degrees[i] for i in self.nodes_]
        timings.append((t.time()-T,"in_out_total_degrees"))

        T=t.time()
        self.strengths=     dict(g.degree(weight="weight"))
        self.nodes__= sorted(g.nodes(), key=lambda x : self.strengths[x])
        self.strengths_=[self.strengths[i] for i in self.nodes_]
        self.in_strengths= dict(g.in_degree(weight="weight"))
        self.in_strengths_=[self.in_strengths[i] for i in self.nodes_]
        self.out_strengths=dict(g.out_degree(weight="weight"))
        self.out_strengths_=[self.out_strengths[i] for i in self.nodes_]
        timings.append((t.time()-T,"in_out_total_strengths"))

        # symmetry measures
        self.asymmetries=asymmetries=[]
        self.disequilibrium=disequilibriums=[]
        self.asymmetries_edge_mean=asymmetries_edge_mean=[]
        self.asymmetries_edge_std=asymmetries_edge_std=[]
        self.disequilibrium_edge_mean=disequilibrium_edge_mean=[]
        self.disequilibrium_edge_std=disequilibrium_edge_std=[]
        for node in self.nodes_:
            if not self.degrees[node]:
                asymmetries.append(0.)
                disequilibriums.append( 0.)
                asymmetries_edge_mean.append(0.)
                asymmetries_edge_std .append(0.)    
                disequilibrium_edge_mean.append(0.)
                disequilibrium_edge_std.append(0.)
            else:
                asymmetries.append(
                    (self.in_degrees[node]-self.out_degrees[node])/self.degrees[node])
                disequilibriums.append( 
                    (self.in_strengths[node]-self.out_strengths[node])/self.strengths[node])
                edge_asymmetries=ea=[]
                edge_disequilibriums=ed=[]
                predecessors=g.predecessors(node)
                successors=g.successors(node)
                for pred in predecessors:
                    if pred in successors:
                        ea.append( 0. )
                        ed.append((g[pred][node]['weight']-g[node][pred]['weight'])/self.strengths[node])
                    else:
                        ea.append( 1. )
                        ed.append(g[pred][node]['weight']/self.strengths[node])
                for suc in successors:
                    if suc in predecessors:
                        pass
                    else:
                        ea.append(-1.)
                        ed.append(-g[node][suc]['weight']/self.strengths[node])
                asymmetries_edge_mean.append(   n.mean(ea))
                asymmetries_edge_std .append(   n.std(ea))  
                disequilibrium_edge_mean.append(n.mean(ed))
                disequilibrium_edge_std.append( n.std(ed)) 


        if "weighted_directed_betweenness" not in exclude:
            T=t.time()
            self.weighted_directed_betweenness=x.betweenness_centrality(g,weight="weight")
            self.weighted_directed_betweenness_=[
  self.weighted_directed_betweenness[i] for i in self.nodes_]
            timings.append((t.time()-T,"weighted_directed_betweenness"))
        if "unweighted_directed_betweenness" not in exclude:
            T=t.time()
            self.unweighted_directed_betweenness=x.betweenness_centrality(g)
            timings.append((t.time()-T,"unweighted_directed_betweenness"))
        if "weighted_undirected_betweenness" not in exclude:
            T=t.time()
            self.weighted_undirected_betweenness=x.betweenness_centrality(gu,weight="weight")
            timings.append((t.time()-T,"weighted_undirected_betweenness"))
        if "unweighted_undirected_betweenness" not in exclude:
            T=t.time()
            self.weighted_undirected_betweenness=x.betweenness_centrality(gu)
            timings.append((t.time()-T,"unweighted_undirected_betweenness"))
        if "wiener" not in exclude:
            T=t.time()
            self.wiener=x.wiener_index(g,weight="weight")
            timings.append((t.time()-T,"weiner"))
        if "closeness" not in exclude:
            T=t.time()
            self.closeness=x.vitality.closeness_vitality(g,weight="weight")
            timings.append((t.time()-T,"closeness"))
        if "transitivity" not in exclude:
            T=t.time()
            self.transitivity=x.transitivity(g)
            timings.append((t.time()-T,"transitivity"))
        if "rich_club" not in exclude:
            T=t.time()
            self.rich_club=x.rich_club_coefficient(gu)
            timings.append((t.time()-T,"rich_club"))

        if "weighted_clustering" not in exclude:
            T=t.time()
            self.weighted_clusterings=x.clustering( network.gu ,weight="weight")
            self.weighted_clusterings_=[self.weighted_clusterings[i] for i in self.nodes_]
            timings.append((t.time()-T,"weighted_clustering"))
        if "clustering" not in exclude:
            T=t.time()
            self.clusterings=x.clustering( network.gu )
            self.clusterings_=[self.clusterings[i] for i in self.clusterings]
            timings.append((t.time()-T,"clustering"))
        if "triangles" not in exclude:
            T=t.time()
            self.triangles=x.triangles(gu)
            timings.append((t.time()-T,"clustering"))
        if "n_weakly_connected_components" not in exclude:
            T=t.time()
            self.n_weakly_connected_components=x.number_weakly_connected_components(g)
            timings.append((t.time()-T,"n_weakly_connected_components"))
        if "n_strongly_connected_components" not in exclude:
            T=t.time()
            self.n_strongly_connected_components=x.number_strongly_connected_components(g)
            timings.append((t.time()-T,"n_strongly_connected_components"))
        T=t.time()
        foo=[i for i in x.connected_component_subgraphs(gu)]
        bar=sorted(foo,key=lambda x: x.number_of_nodes(),reverse=True)
        self.component=c=bar[0]
        timings.append((t.time()-T,"component"))
        T=t.time()
        self.diameter=x.diameter(c)
        self.radius=x.radius(c)
        self.center=x.center(c)
        self.periphery=x.periphery(c)
        timings.append((t.time()-T,"radius_diameter_center_periphery"))
        self.timings=timings

        T=t.time()
        self.n_connected_components=x.number_connected_components(gu)
        nodes=[]
        nodes_components=[foo.nodes() for foo in x.connected_component_subgraphs(gu)][:1]
        for nodes_ in nodes_components: nodes+=nodes_
        self.periphery_=nodes
        self.timings=timings












import music as m, numpy as n, os, pickle
from .networkPartitioning import *
UTILS=m.Utils()
bt=m.BasicTables()
co=m.BasicConverter()
sy=m.Synth()
sy.adsrSetup(A=20,D=20,R=10)
#sy.tab=sy.tables.square
sy2=m.pieces.FourHubsDance()
sy2.sy=sy

class EvolutionMusic:
    """Make music for appreciation of network evolution"""
    def __init__(self,tdir="./evolution/",label="gmaneID",samplerate=44100):
        self.sy=sy
        self.sy2=sy2
        self.m=m
        self.samplerate=samplerate
        self.tdir=tdir
        self.label=label
        self.getMeasures()
        self.makeMusic()
    def makeMusic(self,framerate=4.,renderVideo=True):
        """Make music with 4 frames tempo"""
        self.dur_frame=1/framerate
        self.T=4*self.dur_frame
        # frames per pulse

        # 4 notas por tempo, uma para cada um dos hubs
        agents=self.overall[1][-2].nodes__[-4:]
        # observa os graus dos 4
        #self.degrees=n.array([[nm.degrees.get(agent,0) for nm in self.agents["nm"]] for agent in agents])
        self.degrees=n.array([[nm.strengths.get(agent,0) for nm in self.agents["nm"]] for agent in agents])
        self.n_frames=self.degrees.shape[1]

        total_time=self.n_frames/framerate

        # normaliza cada
        self.amb=12
        self.off=[0,12,24,36][::-1]
        self.degrees_=((n.subtract(self.degrees.T,self.degrees.min(1)) / (self.degrees.max(1)-self.degrees.min(1))).T)
        self.degrees__=(self.degrees_.T+self.off).T

        self.sy2.setupEngine(self.samplerate,total_time,self.T)
        line1=self.sy2.sonicLine2(
   self.degrees_[-1],tmean=.3)
        line2=self.sy2.sonicLine2(
   self.degrees_[-2],tmean=.7,DUR=1)

        sy.adsrSetup(A=5,D=20,S=-20,R=100)
        sy.tab=sy.tables.square
        line4=self.sy2.sonicLine1(
  self.degrees_[0],[0,2,4,6,8,10],ambit=12,rythmic_pattern=[0,0,1,1],f0=440)
        sy.adsrSetup(A=5,D=20,S=-15,R=730)
        # Adicionar vibrato proporcional
        sy.tab=sy.tables.saw
        line3=self.sy2.sonicLine1(
    self.degrees_[1],[0,2,4,6,8,10],ambit=12,rythmic_pattern=[1],f0=110)

        #UTILS.write(line2+line3[:len(line2)]+line4,"mixY.wav")
        llen=len(line4)
        UTILS.write(line1[:llen]+
                    line2[:llen]+
                    line3[:llen]+
                    line4[:llen],self.tdir+"mixY.wav")
        # coloca cada um em uma oitava e escala
        # une os vetores de frequencia
        # junta duração
        #self.midi_notes=n.hstack(self.degrees__.T)

        # fundo feito pelos perifericos
        # buscar alguma sonoridade ruidosa para os intermediarios 

        #self.n_frames=len(self.agents["nm"])
        #self.freqs=co.p2f(110,[[0,7][i%2] for i in range(self.n_frames)])
        #self.freqs=co.p2f(110,self.midi_notes[::4])
        #self.durs=[self.dur_frame]*self.n_frames
        #self.notes=[sy.render(f,d) for f,d in zip(self.freqs,self.durs)]
        #UTILS.write(n.hstack(self.notes),"notes.wav")

        #self.freqs2=co.p2f(110,self.midi_notes[2::12])
        #self.durs2=[self.dur_frame*3]*(self.n_frames//3)
        #self.notes2=[sy.render(f,d) for f, d in zip(self.freqs2, self.durs2)]
        #UTILS.write(n.hstack(self.notes2),"notes2.wav")

        #mix=n.hstack(self.notes)+n.hstack(self.notes2)

        #UTILS.write(mix,"mix.wav")
    def getOverallMeasures(self):
        filenames=os.listdir(self.tdir)
        filenames_=[i for i in filenames if i.endswith(".pickle")]
        filename=[i for i in filenames_ if "overall" in i][0]
        with open(self.tdir+filename,"rb") as f:
                self.overall=pickle.load(f)
    def getMeasures(self):
        self.getOverallMeasures()
        filenames=os.listdir(self.tdir)
        filenames_=[i for i in filenames if i.endswith(".pickle")]
        filenames_=[i for i in filenames_ if i.startswith("im")]
        filenames_.sort()
        agents={"s":[],"is":[],"os":[],
                "d":[],"id":[],"od":[],
                "exc":[],
                "inc":[],
                "nm":[]}
        for filename in filenames_:
            with open(self.tdir+filename,"rb") as f:
                data=pickle.load(f)
                nm=data["nm"]
                agents["nm"].append(nm)
                n_agents=nm.N
                minimum_incidence=data["np"].minimum_incidence
                agents["s"].append(data["np"].sectorialized_agents__)
                agents["is"].append(
                        NetworkPartitioning(nm,minimum_incidence=minimum_incidence,metric="is").sectorialized_agents__ )
                agents["os"].append(
                        NetworkPartitioning(nm,minimum_incidence=minimum_incidence,metric="os").sectorialized_agents__ )
                agents["d"].append(
                        NetworkPartitioning(nm,minimum_incidence=minimum_incidence,metric="d").sectorialized_agents__ )
                agents["id"].append(
                        NetworkPartitioning(nm,minimum_incidence=minimum_incidence,metric="id").sectorialized_agents__ )
                agents["od"].append(
                        NetworkPartitioning(nm,minimum_incidence=minimum_incidence,metric="od").sectorialized_agents__ )
                # pegar os agentes da rede em
                # todos os outros critérios de particionamento simples:
                # d, id, od, is, os
                # e mandar eles para uma função que já calcula
                # os particionamentos compostos
                compound=compoundPartitioning(agents)
                agents["exc"].append(compound["exc"])
                agents["inc"].append(compound["inc"])
        self.agents=agents



from lxml import html
import requests, urllib, os, pickle, logging
import numpy as n

class DownloadGmaneData:
    """Class for downloading Gmane data to local file system

    initialize with a basedir to store email messages
    and with a logging_file to store download information:
    dl=DownloadGmaneData() # OR
    dl=DownloadGmaneData("~/gmane2/","dl.log")

    Methods
    =======
    downloadListIDS(load_local=True) :: retrieves all self.list_ids from Gmane
    downloadListMessages(list_id) :: downloads messages from a GMANE email list with list_id (use self.list_ids[n] if id not at hand)
    getDownloadedLists() :: retrieves all list_ids whose messages have been downloaded
    cleanDownloadedLists() :: cleans lists from empty messages at the end and at the beggining of non-empty messages (empty messages in between are left for coherence).
    downloadedStats() :: raises elementary info about downloaded lists in _BASE_DIR/stats.txt. Generates self.stats and self.lists, the first a dictionary of stats for each list_id and "all" (overall stats). The second a list of each list statistics ordered from the lists with more messages to the ones with the fewest.

    Deprecated methods
    ==================
    correctFilenames() :: corrects the filenames to be 8 digit integers
    """
    def __init__(self,basedir="~/.gmane/",logging_file="download.log"):
        self._BASE_DIR=basedir.replace("~",os.path.expanduser("~"))
        if not os.path.isdir(self._BASE_DIR):
            os.mkdir(self._BASE_DIR)
        self._empty_messages_count=0
        logging.basicConfig(filename=logging_file,format='%(asctime)s: %(message)s',level=logging.DEBUG)
        self.getDownloadedLists()
    def correctFilenames(self):
        """Ensure that all message filenames are 8 digit integers"""
        try:
            self.downloaded_lists
        except AttributeError:
            self.getDownloadedLists()
        for elist in self.downloaded_lists:
            mfiles=os.listdir(self._BASE_DIR+elist)
            mfiles=[i for i in mfiles if i.isdigit()]
            for mfile in mfiles:
                oldname="{}{}/{}".format(self._BASE_DIR,elist,mfile)
                newname="{}{}/{}".format(self._BASE_DIR,elist,"%08d"%int(mfile))
                os.rename(oldname,newname)
    def cleanDownloadedLists(self):
        """Remove empty messages from the beggining or end of each list"""
        try:
            self.downloaded_lists
        except AttributeError:
            self.getDownloadedLists()
        for elist in self.downloaded_lists:
            mfiles=os.listdir(self._BASE_DIR+elist)
            mfiles.sort()
            sizes=[os.path.getsize("{}{}/{}".format(self._BASE_DIR,elist,i)) for i in mfiles]
            sizes_=[i==0 for i in sizes]
            # remove suffix
            try:
                findex=-sizes_[::-1].index(False)
            except ValueError:
                logging.info("no messages for list {}".format(elist))
                for ind, filename in enumerate(mfiles):
                    logging.info("removing file {}/{} of size {}".format(elist,filename,sizes[ind]))
                    os.remove("{}{}/{}".format(self._BASE_DIR,elist,filename))
            else:
                if (findex>0) and (sum(sizes_[findex:])==-findex):
                    # remove todos os arquivos de findex até o final
                    for ind,filename in enumerate(mfiles[findex:]):
                        logging.info("removing file {}/{} of size {}".format(elist,filename,sizes[findex+ind]))
                        os.remove("{}{}/{}".format(self._BASE_DIR,elist,filename))
                    logging.info("file {}/{} have size {}".format(elist,mfiles[findex-1],sizes[findex-1]))
                else:
                    logging.info("no empty suffix for list {}".format(elist))
            # remove prefix
                findex=sizes_.index(False)
                if findex and (sum(sizes_[:findex])==findex):
                    for ind,filename in enumerate(mfiles[:findex]):
                        logging.info("removing file {}/{} of size {}".format(elist,filename,sizes[ind]))
                        os.remove("{}{}/{}".format(self._BASE_DIR,elist,filename))
                    logging.info("file {}/{} have size {}".format(elist,mfiles[findex],sizes[findex]))
                else:
                    logging.info("no empty prefix for list {}".format(elist))

    def getDownloadedLists(self):
        """Retrieves all list_ids from Gmane"""
        dirs=[i for i in os.listdir(self._BASE_DIR) if os.path.isdir(self._BASE_DIR+i)]
        dirs.sort()
        self.downloaded_lists=dirs
    def downloadListIDS(self,count=-1,load_local=True):
        """Downloads at most count GMANE list_ids"""
        if load_local and os.path.isfile(self._BASE_DIR+"lists.pickle"):
            with open(self._BASE_DIR+"lists.pickle","rb") as f:
                self.list_ids=pickle.load(f)
                logging.info("self.list_ids loaded from pickle")
        else:
            page = requests.get('http://gmane.org/lists.php')
            tree = html.fromstring(page.text)
            list_ids = tree.xpath('//a/text()')
            list_ids=list(set(i for i in list_ids if i.startswith("gmane.")))
            self.list_ids=list(list_ids)
            with open(self._BASE_DIR+"lists.pickle","wb") as f:
                pickle.dump(list_ids,f)
            logging.info("self.list_ids created")
    def downloadedStats(self):
        """Raises elementary info about downloaded lists in _BASE_DIR/stats.txt"""
        try:
            self.downloaded_lists
        except AttributeError:
            self.getDownloadedLists()
        stats={}
        sizes_all=[]
        for elist in self.downloaded_lists:
            stats[elist]={}
            mfiles=os.listdir(self._BASE_DIR+elist)
            mfiles.sort()
            sizes=[os.path.getsize("{}{}/{}".format(self._BASE_DIR,elist,i)) for i in mfiles]
            sizes_all+=sizes
            sizes_=[i==0 for i in sizes]
            stats[elist]["n_messages"]=len(mfiles)
            stats[elist]["total_B"]=sum(sizes)
            stats[elist]["average_B"]=n.mean(sizes)
            stats[elist]["std_B"]=n.std(sizes)
            stats[elist]["n_empty"]=sum(sizes_)
        lists=sorted(stats.items(),key=lambda tlist: tlist[1]["n_messages"],reverse=True)
        stats["all"]={}
        stats["all"]["n_lists"]=len(self.downloaded_lists)
        stats["all"]["n_messages"]=len(sizes_all)
        stats["all"]["total_B"]=sum(sizes_all)
        stats["all"]["average_B"]=n.mean(sizes_all)
        stats["all"]["std_B"]=n.std(sizes_all)
        stats["all"]["n_empty"]=sum([i==0 for i in sizes_all])
        with open(self._BASE_DIR+"stats.txt","w") as f:
            f.write("overall downloaded GMANE database stats:\n")
            f.writelines(["{}: {}\n".format(i[0],i[1]) for i in stats['all'].items()])
            f.write("\nstats per list:\n")
            f.write("\n".join([str(i) for i in lists]))
        logging.info("see file {}".format(self._BASE_DIR+"stats.txt"))
        self.stats=stats
        self.lists=lists
    def downloadListMessages(self,list_id=None,start=1,end=10,threshold=5000,replace=False):
        """Download messages form a GMANE email list"""
        self.current_list=list_id
        if not list_id:
            list_id=self.list_ids[0]
        if not os.path.isdir(self._BASE_DIR+list_id):
            os.mkdir(self._BASE_DIR+list_id)
        for i in range(start,end):
            afile = "{}{}/{}".format(self._BASE_DIR,list_id,"%08d"%i)
            self.current_file=afile
            if replace or not os.path.isfile(afile):
                url="http://download.gmane.org/{}/{}/{}".format(list_id,i,i+1)
                self.current_url=url
                urllib.request.urlretrieve(url, afile, self.reporthook)
                if self._empty_messages_count>=threshold:
                    logging.info("download stopped after {} empty messages".format(threshold))
                    for ii in range(i,i-threshold,-1):
                        afile = "{}{}/{}".format(self._BASE_DIR,list_id,str(ii))
                        os.remove(afile)
                    self._empty_messages_count=0
                    with open(self._BASE_DIR+"finishedLists.txt","a") as f:
                        f.write(list_id+"\n")
                    break
            else:
                logging.info("{} exists".format(afile))
    def reporthook(self,a,b,c): 
        if c:
            self._empty_messages_count=0
        else:
            self._empty_messages_count+=1
        logging.info("{} -> {}, {}\nhook {}, {}, {}".format(
                     self.current_url,self.current_file,self._empty_messages_count,a,b,c))



import datetime
def getcharsets(msg):
    charsets = set({})
    for c in msg.get_charsets():
        if c is not None:
            charsets.update([c])
    return charsets
def handleerror2(errmsg, t,charset):
    print(errmsg, charset)
    print("This is the body:",t)
def getBody(msg):
    while msg.is_multipart():
        msg=msg.get_payload()[0]
    try:
        t=msg.get_payload(decode=True)
    except:
        t=msg.get_payload()
    ignore=0
    for charset in getcharsets(msg):
        try:
            if charset=="iso-8859-2:utf-8":
                charset="iso-8859-2"
            elif charset=="us-ascii:utf-8":
                charset="utf-8"
                ignore=1
            elif charset=="unknown-8bit":
                charset="ascii"
                ignore=1
            if ignore:
                t=t.decode(charset,errors="ignore")
            else:
                try:
                    t=t.decode(charset)
                except:
                    try:
                        t=t.decode(charset,errors="ignore")
                    except:
                        if type(t)==type("asid"):
                            return t
                        else:
                            t=t.decode(errors="ignore")

        except LookupError:
            try:
                t=t.decode(charset,errors="ignore")
            except LookupError:
                if type(t)==type("asid"):
                    return t
                else:
                    handleerror2("LookupError for unknown charset:",t,charset)
    if not getcharsets(msg):
        try:
            t=t.decode()
        except:
            try:
                t=t.decode("latin-1")
            except:
                t=t.decode(errors="ignore")
    return t

def cleanText(text):
#    return text
    t=text.splitlines()
    t=[line for line in t if line]
    t_=[]
    for line in t:
        line=line.strip()
        if line.startswith(">"):
            pass
        elif line.startswith("<"):
            pass
        elif "style=" in line:
            pass
        elif "]$" in line:
            pass
        elif "=" in line:
            pass
        elif "INFO" in line:
            pass
        elif (sum([(i in line) for i in ("if","while","for","(",")")])>=3):
            pass
        elif ("FLAGS" in line) and ("=" in line):
            pass
        elif line.endswith(");"):
            pass
        elif line.startswith("$ "):
            pass
        elif line.startswith("return "):
            pass
        elif line.startswith("\./"):
            pass
        elif line.startswith("~/"):
            pass
        elif line.startswith("//"):
            pass
        elif line.startswith(" |"):
            pass
        elif line.startswith("| "):
            pass
        elif line.startswith("On Mon"):
            pass
        elif line.startswith("On Jan"):
            pass
        elif line.startswith("On Tue"):
            pass
        elif line.startswith("On Wed"):
            pass
        elif line.startswith("On Thu"):
            pass
        elif line.startswith("On Fri"):
            pass
        elif line.startswith("On Sat"):
            pass
        elif line.startswith("On Sun"):
            pass
        elif line.endswith("wrote:"):
            pass
        elif sum([line.startswith(i) for i in ("From:","Subject","To","Reply-To:","WARNING")]):
            pass
        # uma palavra soh sem ponto final
        elif len(line.split()) == 1 and line[-1]!=".":
            pass
        # duas palavras separadas, com a primeira letra caixa alta em cada
        elif line.istitle():
            pass
        elif line.startswith("-----BEGIN"):
            pass
        elif line.startswith("Hash: "):
            pass
        elif line.startswith("--"):
            break
        elif line[:4].count("-")>=3:
            break
        elif "----" in line:
            pass
        else:
            t_.append(line)
    t_="\n".join(t_)
    return t_
import string
#allowed=string.punctuation+string.whitespace
allowed=string.punctuation+" \n\t"
def hardClean(text):
    return "".join(c for c in text if ((c.isalnum()) or (c in allowed)))
import urllib
def urifyID(tid):
#    return tid.strip().replace("@","_-AT-_").replace(">","").replace("<","").replace("[","").replace("]","")
    return urllib.parse.quote(tid.strip().replace("@","_-AT-_").replace(">","").replace("<","").replace("[","").replace("]",""))

class FixedOffset(datetime.tzinfo):
    """Fixed offset in minutes east from UTC."""
    def __init__(self, offset, name):
        self.__offset = datetime.timedelta(minutes=offset)
        self.__name = name

    def utcoffset(self, dt):
        return self.__offset

    def tzname(self, dt):
        return self.__name

    def dst(self, dt):
        return ZERO



import networkx as x
class InteractionNetwork:
    """Delivers the interaction network of email data structures

    Network is directed and weighted.

    Usage
    =====
    Initialize with a MessageDataStructures() instance.
    """
    def __init__(self,listStructures=None):
        if not listStructures:
            g.listDataStructures()
        g=x.DiGraph()
        self.makeGraph(listStructures, g)
    def makeGraph(self,listStructures, g):
        """Make the interaction digraph"""
        for tid in listStructures.message_ids:
            m=listStructures.messages[tid]
            if m[0] in g.nodes(): # author
                g.node[m[0]]["weight"]+=1
            else:
                g.add_node(m[0],weight=1.)
            if m[1]:
                if m[1] in listStructures.message_ids:
                    m0=listStructures.messages[m[1]]
                    if g.has_edge(m0[0],m[0]):
                        g[m0[0]][m[0]]["weight"]+=1
                    else:
                        if m0[0] not in g.nodes():
                            g.add_node(m0[0],weight=0.)
                        g.add_edge(m0[0], m[0], weight=1.)
        # no selfloops, as interest is on interaction
        g.remove_edges_from(g.selfloop_edges())
        self.g=g
        self.gu=g.to_undirected()



import dateutil.parser, parsedatetime as pdt, pytz
import percolation as P, gmane as G
parser=pdt.Calendar()
c=P.check
class ListDataStructures:
    """Basic datastructures driven from Gmane email messages.

    Usage
    =====
    Initialize with a LoadMessages() instance.
    If not interested in the textual content, use test="no"
    to avoid the overhead.

    Attributes
    ==========
    After initialization of class, you get:
    self.messages a dict for infos by message_id as key
    self.message_ids a cronologically ordered list of message_id s
    self.empty_ids list of message_id s of empty messages
    self.author_messages  messages by author email as keys
    self.n_messages # number of messages
    self.n_authors # number of authors 
    self.responses for accessing responses to a message by the message_id
    self.raw_clean_autors keep both the original author field and cleaned email address used as author ID
    self.spurious_author should keep only messages with no true sender
    self.raw_clean_dates keep both the original date field and compatibilized with dateutil.parser

    Precedent implementation variable names
    =======================================
    For compatibility with algorithms that used previous implementations
    it is worth keeping the track of the variable names.
    This is the correspondence:
    mm changed to messages
    ids to message_ids
    aa to author_messages
    N to n_messages
    ids_r to responses
    vz to spurious_empty_ids
    """


    def __init__(self, messagesLoaded=None,text="yes"):
        if "messages" in dir(messagesLoaded):
            messagesLoaded=messagesLoaded.messages
        self.messagesLoaded=messagesLoaded
        self.messages=messages={}
        self.message_ids=message_ids=[]
        self.author_messages=author_messages={}
        self.n_messages=len(messagesLoaded)
        self.responses=responses={}
        self.raw_clean_authors=raw_clean_authors=[]
        self.raw_clean_dates=raw_clean_dates=[]
        self.raw_clean_references=raw_clean_references=[]
        #self.spurious_empty_ids=spurious_empty_ids=[]
        self.spurious_authors=spurious_authors=[]
        self.n_empty=n_empty=0#messagesLoaded.n_empty
        self.bad_text_message_ids=[]
        mcount=0
        for message in messagesLoaded:
            if not message.keys(): # if message is empty
                #spurious_empty_ids.append(i)
                self.n_empty+=1
            else:
                mcount+=1
                if mcount%300==0: c("struct +300: {}".format(mcount))
                author_=message['from']
                if "replace" not in dir(author_):
                    spurious_authors.append(message)
                    continue

                author=author_.replace('"','')
                author=author.split("<")[-1][:-1]
                if " " in author: 
                    author=author.split(" ")[0]
                raw_clean_authors.append((author_,author))
                if author not in author_messages:
                    author_messages[author]=[]
                date_=message['date']
                try:
                    date = date_.split(" (")[0]
                except:
                    continue
                if date.split(" ")[-1].islower():
                    date=date.replace(date.split(" ")[-1],date.split(" ")[-1].upper())
                if date.split(" ")[-1].isupper() and date.split(" ")[-1].isalpha():
                    date=date.replace(date.split(" ")[-1],"")
                #date=date.replace("GMT","")
                #date=date.replace(" CST","")
                #date=date.replace(" CDT","")
                date=date.replace("(KST)","")
                date=date.replace("Thur","Thu")
                date=date.replace("--","-")
                if "+-" in date:
                    date=date.split("+-")[0][:-1]
                if "-" in date and len(date.split("-")[-1])==3:
                    date=date+"0"
                if "+" in date and len(date.split("+")[-1])==3:
                    date=date+"0"
                if "\n" in date:
                    date=date.split("\n")[0]
                if date.startswith("So, "):
                    date=date[4:]
                date=dateutil.parser.parse(date)
                if date.tzinfo==None: # colocando localizador em que não tem, para poder comparar
                    date=pytz.UTC.localize(date)
                try:
                    aa=date.utcoffset()
                    aa=date.isoformat()
                except:
                    date=date.replace(tzinfo=None)
                    date=pytz.UTC.localize(date)
                raw_clean_dates.append((date_,date))
                if message['references']:
                    try:
                        id_ant=message['references'].split('\t')[-1]
                        id_ant=id_ant.split(' ')[-1]
                    except:
                        print("AQUI AQUI")
                        continue
                else:
                    id_ant=None
                if id_ant:
                    if id_ant not in responses.keys():
                        responses[id_ant]=[]
                    responses[id_ant].append( (author,message["message-id"],date) )
                    raw_clean_references.append((message['references'],id_ant))
                if text=="no":
                    messages[message["message-id"]]=(author,id_ant,date)
                elif text=="yes":
#                    try:
                    t=G.utils.getBody(message) # AQUI TTM
                    t_=G.utils.cleanText(t) # AQUI TTM
                    t_=G.utils.hardClean(t_) # AQUI TTM
                    messages[message["message-id"]]=(author,id_ant,date,t_)
#                    except:
#                        messages[message["message-id"]]=(author,id_ant,date,"")
#                        self.bad_text_message_ids.append(message["message-id"])
                else:
                    raise TypeError("argument text accepts only 'yes' and 'no'values")
                author_messages[author].append( (message["message-id"], id_ant, date)  )
                message_ids.append(message['message-id'])
                #self.n_empty=len(spurious_empty_ids)
                self.n_authors=len(author_messages)
